#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include "event.h"
#include "event_constants.h"

/* instead use M_PI or M_2PI defined in math.h */

#define TWOPI  2* 3.1415926535897932384626434
#define ISIP_TRUE 1
#define ISIP_FALSE 0

int main(int argc, char **argv)
{

  int merge(int npro, FILE* event_file, char* file_string);
  int sort( char* file_string);

  double start_time;
  double end_time;
  FILE* event_file;
  FILE* local_event_file;
  char file_string[MAX_STRING_LENGTH];
  char local_file_string[MAX_STRING_LENGTH];
    
    int myrank,npro,m,p, prev,next,numb,size;
    double starttime, total_time,maxtime;
    float *data; /* The input data is stored here*/
    float *out;  /* The output is stored here */
    float *local_data; /* This is allocated for each processor input */
    float *local_out; /* This is allocated for each processor output */
    int i;
    double cpu_time_d,cpu_time_f;
    void dft_real_cc() ;
    
    /* To Initialize MPI */
    
    MPI_Init(&argc , &argv);
    MPI_Comm_rank (MPI_COMM_WORLD,&myrank);
    MPI_Comm_size  (MPI_COMM_WORLD,&npro);
    
    if( argc != 3) {
      if( myrank == 0 ) {
	printf("USAGE: %s <number_of_points> <event file name>\n",argv[0] );
      }
      MPI_Finalize();
      return( ISIP_FALSE );
    }

    strcpy(&file_string, argv[2]);
    sprintf(local_file_string, "%s.%d", file_string, myrank);
    local_event_file = fopen(local_file_string,"w");
    
    if ( myrank == 0 ) {
      event_file = fopen(file_string,"w");
    }

    m= size % npro;
    if (m!= 0) { 
      if(myrank == 0)
	printf("SORRY NOT EQUIPPED TO HANDLE THIS\n");
      MPI_Finalize();
      exit(1);
    } 
    
    /*
     * retrieve number of data points to use from command line
     */
    size = atoi( argv[1] );

    /* Gives number of elements that have to be sent to each processor. */    
    numb = size/npro; 
    
    data  = (float *)malloc (size * sizeof (float));
	
    if( myrank ==0)
    {
	for (i=0; i<size; i++)
	{ 
	    data[i] =i;
	}       

	/* This data is local only to the zero processor 
	   this cannot be acessed by other processors */
	
	out  = (float *)malloc (size * sizeof (float)); 
    }

    /* mark the initial time of each process */
    MPI_Barrier( MPI_COMM_WORLD );
    start_time= MPI_Wtime();

    /* This send's same copy of data to all the processors. */
    if( npro > 1 ) {
      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank,0);
      }

      MPI_Bcast(data,size,MPI_FLOAT,0,MPI_COMM_WORLD); 

      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    }

    
    /* This allocates data & local_out on all the processors 
       to store the the output. */
    
    local_out = (float *)malloc (numb * sizeof (float)); 
    
    /* The routine takes local_data input and performs DFT and
       passes local_out as output , which is smaller than the 
       "data" size which is the input vector.*/

    dft_real_cc(local_out,data,size,npro,myrank);
    
    /* This collects all the local vectors and places the sum in the 
       out which is local to the zero processor */
    
  if( npro > 1 ) { 

    if ( myrank == 0 ) {
      fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
	      MPI_Wtime()-start_time, myrank, ALL_NODES);
    }
    else {
      fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
	      MPI_Wtime()-start_time, myrank, 0);
    }
      
    MPI_Gather(local_out, numb, MPI_FLOAT, out, numb, MPI_FLOAT,
	         0,MPI_COMM_WORLD);

    fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	    MPI_Wtime()-start_time, myrank, myrank);
  } 

  fprintf(local_event_file,"%d %f %d %d\n", STOP_EVENT, 
	  MPI_Wtime()-start_time, myrank, myrank );
    
  fclose(local_event_file);      
  
  if ( myrank == 0 ) {
    merge(npro, event_file, file_string);
    sort( file_string );
  }
       

#ifdef Test

    if(myrank == 0)
    {
	for(i=0;i<size;i++)
	    printf(" out[%d] = %f \n",i,out[i]);
    }

#endif
    
    MPI_Finalize();

    /* exit gracefully */
    return (ISIP_TRUE);
    
}



#define TWOPI  2* 3.1415926535897932384626434
void dft_real_cc(float *out_a, float *in_a, int N_d, int npro, int myrank) 
{
    int k,n,no_ele;
    float sumr,sumi,arg;
    int offset;

    /* Gives no. of data points of the total data being copmuted */
    no_ele = N_d/npro;
    offset = no_ele*myrank;

    for ( k=0; k<no_ele; k++) {
	
	/* declare local accumulators*/
	sumr = 0.0;
	sumi = 0.0;
	
	/* offset is used as the the data on each processor is needs to be 
	   multiplied by different n's */

	for( n=0; n<N_d; n++) 
	{
	    arg = TWOPI * (float)n * (float)offset / (float)N_d;
	    sumr += in_a[n]*cos(arg);
	    sumi += -1.0*in_a[n]* sin(arg);
	}
	offset++;
	out_a[k] = sumr*sumr;
	out_a[k] += sumi*sumi;

    }
       
}


