/* file: merge_and_sort.c  */


/* include files */

#include <stdio.h>
#include <stdlib.h>
#include "event.h"
#include "event_constants.h"

/* method: merge_and_sort
//
// This routine merges npro files into a single ordered file.
// the ordering is based on the time stamp which is the second string in
// the file.
//
// arguments:
// int npro :(input) number of files to open
// FILE* event_file : (input) file to write the sorted data to.
// char* file_string : (input) the base filename of the data files
//
// return: a logical_1 value indicating status
//
// merges a number of files and sorts them according to the first number on
// each line of the file
*/

#define ISIP_TRUE 1
#define ISIP_FALSE 0

int merge( int npro, FILE* event_file, char* file_string ) {
  FILE* current_file;
  int index;
  char next_line[MAX_STRING_LENGTH];
  char current_file_string[MAX_STRING_LENGTH];
  
  for ( index = 0; index < npro; index++) {

    sprintf(current_file_string, "%s.%d", file_string, index);
    current_file = fopen(current_file_string,"r");

    while (fgets(next_line,MAX_STRING_LENGTH,current_file) != NULL) {
      fprintf(event_file,"%s",next_line);
    }

    fclose(current_file);

  }
  fclose(event_file);
  
  return(ISIP_TRUE);

}



int sort( char* file_string ) {

  int selection_sort(struct event_data_rec event_data[], int num_of_records);
  FILE* current_file;
  FILE* temp_file;
  int index=0;
  int num_of_records=0;
  struct event_data_rec event_data[MAX_NUM_RECORDS];
  
  current_file = fopen(file_string, "r");

  while (fscanf(current_file, "%d %f %d %d\n",
		&event_data[index].event,&event_data[index].time_stamp,
		&event_data[index].send_node, &event_data[index].recv_node) 
	 != EOF) {
    index++;
    num_of_records++;
  }
  
  fclose(current_file);

  selection_sort(event_data, num_of_records);

  current_file = fopen(file_string, "w");

  for (index = 0; index < num_of_records; index++){
    fprintf(current_file, "%d %f %d %d\n",event_data[index].event,
	    event_data[index].time_stamp,event_data[index].send_node, 
	    event_data[index].recv_node);
  }
  
  fclose( current_file ); 

  return(ISIP_TRUE);
}

int selection_sort(struct event_data_rec event_data[], int num_of_records) {

  int index1, index2;
  int smallest_index;
  int swap(struct event_data_rec event_data[], int current, int low);

  for ( index1 = 0; index1 < num_of_records - 1; index1++) {
    
    smallest_index = index1;
    
    for ( index2 = index1 + 1; index2 < num_of_records; index2++) {
      if (event_data[index2].time_stamp < 
	  event_data[smallest_index].time_stamp) {
	smallest_index = index2;
      }
    }
    
    if (index1 != smallest_index) {
      swap(event_data, index1, smallest_index);
    }
  }
  
  return (ISIP_TRUE);
}
    

int swap(struct event_data_rec event_data[], int current, int low) {

  struct event_data_rec temp_rec;

  temp_rec = event_data[current];
  event_data[current] = event_data[low];
  event_data[low] = temp_rec;

  return (ISIP_TRUE);
}
