/* file: ditf.c*/

/* include files*/
#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include "event.h"
#include "event_constants.h"

/* define pi*/
#define pi 3.14159265359

#define ISIP_TRUE 1
#define ISIP_FALSE 0

#define BASE_NODE 0

int main(int argc, char* argv[]) {
    int numb;
    int myrank,npro;
    double* input;
    double* real_a;
    double* imag_a;
    double* local_real_a;
    double* local_imag_a;
    int i;
    void ditf_calc(double* input, int v, int gamma, int two_v, int n, int N_d,
		   double* real_result_ptr, double* imag_result_ptr,
		   int end_stage);
    void dit(double* input, int v, int gamma, int two_v, int n, int N_d,
	     double* real_result_ptr, double* imag_result_ptr);
    int merge(int npro, FILE* event_file, char* file_string);
    int sort( char* file_string);
    
    double start_time;
    double end_time;
    FILE* event_file;
    char file_string[MAX_STRING_LENGTH];
    char local_file_string[MAX_STRING_LENGTH];
    FILE* local_event_file;

    int N_d;
    int v;	  /* assign v = log2(N_d)*/
    int two_v;		  /* two_v = 2^v */
    int end_stage;
    int offset;

    MPI_Init(&argc , &argv);
    MPI_Comm_rank (MPI_COMM_WORLD,&myrank);
    MPI_Comm_size  (MPI_COMM_WORLD,&npro);
    

    if( argc != 3) {
      if( myrank == 0 ) {
	printf("USAGE: %s <number_of_points> <event file name>\n",argv[0] );
      }
      MPI_Finalize();
      return( ISIP_FALSE );
    }

    strcpy(&file_string, argv[2]);
    sprintf(local_file_string, "%s.%d", file_string, myrank);
    local_event_file = fopen(local_file_string,"w");
    
    if ( myrank == 0 ) {
      event_file = fopen(file_string,"w");
    }
    
    N_d = atoi(argv[1]);
    v = log10(N_d)/log10(2);
    two_v = N_d;
    end_stage = v/2;
    
    input = (double*) malloc(sizeof(double) * N_d);

    if (myrank == 0) {
	real_a = (double*) malloc(sizeof(double)*N_d);
	imag_a = (double*) malloc(sizeof(double)*N_d);
    }


    if (myrank==0) {
	for (i=0;i<N_d;i++) {
	    input[i] = i;
	}
    }

    /* mark the initial time of each process */
    MPI_Barrier( MPI_COMM_WORLD );
    start_time = MPI_Wtime();  

    if( npro > 1 ) {
      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank, BASE_NODE);
      }

      MPI_Bcast( input, N_d, MPI_DOUBLE, BASE_NODE,
		 MPI_COMM_WORLD );
      
      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    }
    
    if ( N_d%(2*npro) ==  0 ) {
      numb = N_d/(2*npro);
    }
    else {
      numb = N_d/npro;
    }
   
    local_real_a = (double*) malloc(numb * sizeof(double));
    local_imag_a = (double*) malloc(numb * sizeof(double));
    offset = numb * myrank;

    for(i=offset; i<(offset + numb); i++) {
	ditf_calc(input,v,i,two_v,i,N_d,&local_real_a[i-offset],
		  &local_imag_a[i-offset],end_stage);
    }
    
    if ( 1 < npro ) {
      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, BASE_NODE);
      }
      
      MPI_Gather(local_real_a, numb, MPI_DOUBLE, real_a, numb, MPI_DOUBLE,
		 BASE_NODE,MPI_COMM_WORLD);
      
      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    
      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, BASE_NODE);
      }
      
      MPI_Gather(local_imag_a, numb, MPI_DOUBLE, imag_a, numb, MPI_DOUBLE,
		 BASE_NODE,MPI_COMM_WORLD);
      
      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    } 
    
    if ( myrank == 0 ) {
      if ( N_d%(2*npro) ==  0 ) {
	for (i = 1;i <= N_d/2;i++){
	  imag_a[N_d - i] = -imag_a[i];
	  real_a[N_d - i] = real_a[i];
	}
	imag_a[N_d/2] = imag_a[0] = 0.0;
      }
    }

#ifdef Debug
    
    if(myrank == 0) {
	for(i=0;i<N_d;i++)
	    printf(" out[%d] = %f  +j %f \n",i,real_a[i],imag_a[i]);
    }
#endif

    if ( myrank == 0 ) {
      end_time = MPI_Wtime()-start_time;
    }
      fprintf(local_event_file,"%d %f %d %d\n", STOP_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);

    fclose(local_event_file);
    
    if ( myrank == 0 ) {
      merge(npro, event_file, file_string);
      sort( file_string );
    }
    
    MPI_Finalize();
    /* exit gracefully
     */
    return ISIP_TRUE;
    
}


/* subroutine to calculate the dft using the recursive decimation-in-time*/
/* algorithm described by Ali Saidi in the ICASSP Proceedings of 1994.*/
/* this DIT should not be used alone because it outputs values in a special*/
/* form to be consistent with the DITF*/
/* input: input - array of sample points*/
/*   		 v - stage currently in*/
/*			 gamma - node of the stage*/
/*			 two_v - 2^v*/
/*			 n - sample point currently calculating*/
/*			 N_d - number of samples*/
/*			 real_result_ptr - pointer to the real part of the calculated result*/
/*			 imag_result_ptr - pointer to the imag part of the calculated result*/
/* output: None*/
/* description: this function uses a recursive call to itself to calculate the*/
/*		 coefficients of the dft at gamma node at stage v for the nth*/
/*		 sample point using the following recursion.*/
/*              if gamma = v = 0*/
/*		       g'[gamma,v](n) = x(n)*/
/*              if gamma < 2^(v-1)*/
/*     	         g'[gamma,v](n) = g'[gamma,v-1](n) +
		 g'[gamma,v-1]((n+N_d/(2^v))%N_d)W^(gamma/2^(v-1))
		 else
		 lambda = gamma - 2^(v-1)
		 g'[gamma,v](n) = g'[gamma,v-1](n) -
		 g'[gamma,v-1]((n+N_d/(2^v))%N_d)W^(lambda/2^(v-1))
		 
		 */
void dit(double* input, int v, int gamma, int two_v, int n, int N_d,
	 double* real_result_ptr,	double* imag_result_ptr)
{
    /* define local variables
     */
    double real1 = 0.0,
	real2 = 0.0,
	real3 = 0.0,
	imag1 = 0.0,
	imag2 = 0.0,
	imag3 = 0.0;
    
    /* define variables that will be used often to avoid repeat calculations
     */
    int v_minus = v-1;
    int half_two_v = two_v/2;
    
    /* if v = gamma = 0
       assign return x[n]
       */
    if ((v == 0) && (gamma == 0)) {
	*real_result_ptr = input[n];
	*imag_result_ptr = (double) 0;
    }
    
    /* if gamma <= 2^(v-1) do recursion
     */
    else if (gamma < half_two_v) {
	
	/* calculate complex multiplier
	 */
	double arg = ((double)gamma / (double)half_two_v * pi);
	double x1 = cos(arg);
	double x2 = sin(arg);
	
	/* recursively compute the DIT
	 */
	dit(input,v_minus,gamma,half_two_v,n,N_d,&real1,&imag1);
	dit(input,v_minus,gamma,half_two_v,(n+N_d/two_v)%N_d,N_d,&real2,&imag2);
	
	/* do complex multiplication and store results
	 */
	real3 = (real2*x1) + (imag2*x2);
	imag3 = (imag2*x1) - (real2*x2);
	*real_result_ptr = real1 + real3;
	*imag_result_ptr = imag1 + imag3;
    }
    
    /* if gamma > 2^(v-1) then do recursion
     */
    else {
	
/*	 assign lambda = gamma - 2^(v-1)*/
/*	*/
	int lambda = gamma - half_two_v;
	
/*	 compute complex multiplier*/
/*	*/
	double arg = ((double)lambda / (double)half_two_v * pi);
	double x1 = cos(arg);
	double x2 = sin(arg);
	
/*	 recursively compute DIT*/
/*	*/
	dit(input,v_minus,lambda,half_two_v,n,N_d,&real1,&imag1);
	dit(input,v_minus,lambda,half_two_v,(n+N_d/two_v)%N_d,N_d,&real2,&imag2);
	
/*	 do complex multiplication and store results*/
	real3 = (real2 * x1) + (imag2 * x2);
	imag3 = (imag2 * x1) - (real2 * x2);
	*real_result_ptr = real1 - real3;
	*imag_result_ptr = imag1 - imag3;
    }
}



/* subroutine to calculate the dft using the decimation-in-time-frequency*/
/* algorithm described by Ali Saidi in the ICASSP Proceedings of 1994.*/
/* input: input - array of sample points*/
/*   		 v - stage currently in*/
/*			 gamma - node of the stage*/
/*			 two_v - 2^v*/
/*			 n - sample point currently calculating*/
/*			 N_d - number of samples*/
/*			 real_result_ptr - pointer to the real part of the calculated result*/
/*			 imag_result_ptr - pointer to the imag part of the calculated result*/
/*			 end_stage - the stage at which to transition to the recursive DIT*/
/* output: None*/
/* description: this function uses a recursive call to itself to calculate the*/
/* 	 coefficients of the dft at the gamma node of stage v for the nth*/
/*	 sample point using the following recursion.*/
/*         if gamma = v = 0*/
/*		g[gamma,v](n) = x(n)*/
/*         if gamma < 2^(v-1)*/
/*		g[gamma,v](n) = g[gamma,v-1](n) +*/
/*		 g[gamma,v-1]((n+N_d/(2^v))%N_d)*/
/*	 else*/
/*        	 lambda = gamma - 2^(v-1)*/
/*	         g[gamma,v](n) = (g[gamma,v-1](n) -*/
/*		  g[gamma,v-1]((n+N_d/(2^v))%N_d))W^(n*2^(v-1))*/
/*	this is repeated until v = (log2(N_d))/2*/
/*	at that point the remaining iterations are carried out by the*/
/*	recursive decimation-in-time algorithm and the result is*/
/*	multiplied by W^(gamma*n)*/
void ditf_calc(double* input, int v, int gamma, int two_v, int n, int N_d,
	       double* real_result_ptr,	double* imag_result_ptr, int end_stage)
{
    
/*     local variables*/
/*    */
    
    double real1 = 0.0,
	real2 = 0.0,
	imag1 = 0.0,
	imag2 = 0.0;
    double x1,x2;
    
    
/*     determine if it is time to switch to the DIT portion of the recursion*/
/*    */
    if (v > end_stage) {
	
/*	  assign variables which are used multiple times to avoid recalculation*/
/*	*/
	int v_minus = v - 1;
	int half_two_v = two_v/2;
	
/*	 if v = gamma = 0, return x[n]*/
/*	*/
	if ((v == 0) && (gamma == 0)) {
	    *real_result_ptr = input[n];
	    *imag_result_ptr = (double) 0;
	}
	
/*	 if gamma <= 2^(v-1) do recursion*/
/*	*/
	else if (gamma < (half_two_v)) {
	    ditf_calc(input,v_minus,gamma,half_two_v,n,N_d,&real1,&imag1,end_stage);
	    ditf_calc(input,v_minus,gamma,half_two_v,(n+N_d/(two_v))%N_d,N_d,
		      &real2,&imag2,end_stage);
	    *real_result_ptr = real1 + real2;
	    *imag_result_ptr = imag1 + imag2;
	}
	
/*	 if gamma > 2^(v-1) do recursion*/
/*	*/
	else {
	    int lambda = gamma - half_two_v;
	    double arg = ((n * half_two_v * 2 * pi) / N_d);
	    ditf_calc(input,v_minus,lambda,half_two_v,n,N_d,&real1,&imag1,end_stage);
	    ditf_calc(input,v_minus,lambda,half_two_v,(n + N_d/(two_v))%N_d,N_d,
		      &real2,&imag2,end_stage);
	    real1 = real1-real2;
	    imag1 = imag1-imag2;
	    
/*	     calculate W^(n*2^(v-1))*/
/*	    */
	    x1 = cos(arg);
	    x2 = sin(arg);
	    
/*	     do complex multiplication*/
/*	    */
	    *real_result_ptr = (real1 * x1) + (imag1 * x2);
	    *imag_result_ptr = (imag1 * x1) - (real1 * x2);
	    
	}
    }
    
/*     switch to the DIT portion of the recursion*/
/*    */
    else {
	
/*	 calculate complex conversion factor*/
/*	*/
	double arg = ((double)gamma * n * 2 * pi / (double)N_d);
	double x1 = cos(arg);
	double x2 = sin(arg);
	
/*	 call the DIT recursion and multiply by conversion factor*/
/*	*/
	dit(input,v,gamma,two_v,n,N_d,&real1,&imag1);
	*real_result_ptr = real1 * x1 + imag1 * x2;
	*imag_result_ptr = imag1 * x1 - real1 * x2;
    }
    
    
} /* end ditf_calc*/
