/* file: ditf.c */


/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>


/* define pi */

#define pi 3.14159265359

/* program: ditf */

/* arguments: */
/* 	N: number of samples */
/*		input: input array of N samples */
/*		real_a : array of size N to store real coeff's in. */
/*		imag_a : array of size N to store imag coeff's in. */


/*return:  logical value indicating status */

#define N 131072
#define Debug
int myrank;
int npro;
int npro_this_level;
int npro_offset;
MPI_Status status;

int main(int argc, char* argv[]) {
  
  float input[N];
  float real_a[N];
  float imag_a[N];
  int i;
  void ditf_calc(float* input, int v, int gamma, int two_v, int n, int N_d,
		 float* real_result_ptr, float* imag_result_ptr,
		 int end_stage);
  void dit(float* input, int v, int gamma, int two_v, int n, int N_d,
	   float* real_result_ptr, float* imag_result_ptr);
  
  int N_d=N;
  int v=log10(N)/log10(2);	   /* assign v = log2(N_d) */
  int two_v = N;		   /* two_v = 2^v  */
  int end_stage = v/2;
  int offset;
  int clock_t;
  int niter;
  
  MPI_Init(&argc , &argv);
  MPI_Comm_rank (MPI_COMM_WORLD,&myrank);
  MPI_Comm_size  (MPI_COMM_WORLD,&npro);
  if (myrank == 0) {
    clock_t = (int) clock();
  }
  
  
  for (niter = 0; niter <1000; niter++) {  
    npro_this_level = npro;
    npro_offset = 0;
    if (myrank == 0) {
      for (i=0;i<N_d;i++) {
	input[i] = i;
      }
    }
    
    
    MPI_Bcast(input, N_d, MPI_FLOAT, 0, MPI_COMM_WORLD);
    
    if ((myrank < npro/2) || (npro == 1)) {
      for(i=0; i<N_d/4; i++) {
	ditf_calc(input,v,i,two_v,i,N_d,&real_a[i],
		  &imag_a[i],end_stage);
      }
    }
    if ((npro/2 <= myrank) || (npro == 1)) {
      for(i=N_d/4; i <= N_d/2; i++) {
	npro_offset = npro/2;
	ditf_calc(input,v,i,two_v,i,N_d,&real_a[i],
		  &imag_a[i],end_stage);
      }
    }
    
    if ((4<=N_d) && (npro != 1)) {
      if (myrank == 0) {
	MPI_Recv(&real_a[N_d/4], N_d/4+1, MPI_FLOAT, npro/2, 0, MPI_COMM_WORLD, &status);
	MPI_Recv(&imag_a[N_d/4], N_d/4+1, MPI_FLOAT, npro/2, 1, MPI_COMM_WORLD, &status);
      }
      else if (myrank == npro/2) {
	MPI_Send(&real_a[N_d/4], N_d/4+1, MPI_FLOAT, 0, 0, MPI_COMM_WORLD);
	MPI_Send(&imag_a[N_d/4], N_d/4+1, MPI_FLOAT, 0, 1, MPI_COMM_WORLD);
      }
    }
    
    if (myrank == 0) {
      for (i = 1; i<N_d/2; i++) {
	real_a[N_d/2+i] = real_a[N_d/2-i];
	imag_a[N_d/2+i] = -imag_a[N_d/2-i];
      }
    }
    
    if(myrank == 0) {
      for(i=0;i<N_d;i++)
	printf(" out[%d] = %f  +j %f \n",i,real_a[i],imag_a[i]);
    }
  }
  if (myrank == 0) {
    clock_t = (int) clock();
    clock_t = clock_t/1000;
    printf("total time = %d\n",clock_t);
  }  

  MPI_Finalize();
  /* exit gracefully */
  
  return 1;
  
}


/* subroutine to calculate the dft using the recursive decimation-in-time */
/* algorithm described by Ali Saidi in the ICASSP Proceedings of 1994. */
/* this DIT should not be used alone because it outputs values in a special */
/* form to be consistent with the DITF */

/* input: input - array of sample points */
/*   		 v - stage currently in */
/*			 gamma - node of the stage */
/*			 two_v - 2^v */
/*			 n - sample point currently calculating */
/*			 N_d - number of samples */
/*			 real_result_ptr - pointer to the real part of the calculated result */
/*			 imag_result_ptr - pointer to the imag part of the calculated result */

/* output: None */

/* description: this function uses a recursive call to itself to calculate the */
/*		 coefficients of the dft at gamma node at stage v for the nth */
/*		 sample point using the following recursion. */
/*              if gamma = v = 0 */
/*		       g'[gamma,v](n) = x(n) */
/*              if gamma < 2^(v-1) */
/*     	         g'[gamma,v](n) = g'[gamma,v-1](n) + */
/*		  g'[gamma,v-1]((n+N_d/(2^v))%N_d)W^(gamma/2^(v-1)) */
/*		else */
/*		  lambda = gamma - 2^(v-1) */
/*		  g'[gamma,v](n) = g'[gamma,v-1](n) - */
/*		  g'[gamma,v-1]((n+N_d/(2^v))%N_d)W^(lambda/2^(v-1)) */


void dit(float* input, int v, int gamma, int two_v, int n, int N_d,
	 float* real_result_ptr,	float* imag_result_ptr)
{
  /*  define local variables */
  
  float real1 = 0.0,
    real2 = 0.0,
    real3 = 0.0,
    imag1 = 0.0,
    imag2 = 0.0,
    imag3 = 0.0;
  
  /*  define variables that will be used often to avoid repeat calculations */
  
  int v_minus = v-1;
  int half_two_v = two_v/2;
  
  /*  if v = gamma = 0 */
  /* assign return x[n] */
  
  if ((v == 0) && (gamma == 0)) {
    *real_result_ptr = input[n];
    *imag_result_ptr = (float) 0;
  }
  
  /*  if gamma <= 2^(v-1) do recursion */
  
  else if (gamma < half_two_v) {
    
    /* calculate complex multiplier */
    
    float arg = ((float)gamma / (float)half_two_v * pi);
    float x1 = cos(arg);
    float x2 = sin(arg);
    
    /* recursively compute the DIT */
    if (npro_this_level/2 <= 1) {
      dit(input,v_minus,gamma,half_two_v,n,N_d,&real1,&imag1);
      dit(input,v_minus,gamma,half_two_v,(n+N_d/two_v)%N_d,N_d,&real2,&imag2);
    }
    else {
      npro_this_level = npro_this_level/2;
      if ((myrank < (npro_offset + npro_this_level/2))|| (npro_this_level <= 1)) {
	npro_offset = npro_offset;
	dit(input,v_minus,gamma,half_two_v,n,N_d,&real1,&imag1);
	npro_offset = npro_offset;
	if ((1 < npro_this_level) && (myrank == npro_offset)) {
	  MPI_Recv(&real2, 1, MPI_FLOAT, (npro_offset + npro_this_level/2), 0, 
		   MPI_COMM_WORLD, &status);
	  MPI_Recv(&imag2, 1, MPI_FLOAT, (npro_offset + npro_this_level/2), 1, 
		   MPI_COMM_WORLD,&status);
	}
      }
      else if (((npro_offset + npro_this_level/2) <= myrank) || (npro_this_level <= 1)){
	npro_offset = (npro_offset + npro_this_level/2);
	dit(input,v_minus,gamma,half_two_v,(n+N_d/two_v)%N_d,N_d,&real2,&imag2);
	npro_offset = (npro_offset - npro_this_level/2);
	if ((1 < npro_this_level) && (myrank == npro_offset + npro_this_level/2)) {
	  MPI_Send(&real2, 1, MPI_FLOAT, npro_offset, 0, MPI_COMM_WORLD);
	  MPI_Send(&imag2, 1, MPI_FLOAT, npro_offset, 1, MPI_COMM_WORLD);
	}
      }
      npro_this_level = npro_this_level*2;
    }
    
    /* do complex multiplication and store results */
    
    real3 = (real2*x1) + (imag2*x2);
    imag3 = (imag2*x1) - (real2*x2);
    *real_result_ptr = real1 + real3;
    *imag_result_ptr = imag1 + imag3;
  }
  
  /*  if gamma > 2^(v-1) then do recursion */
  
  else {
    
    /* assign lambda = gamma - 2^(v-1) */
    
    int lambda = gamma - half_two_v;
    
    /* compute complex multiplier */
    
    float arg = ((float)lambda / (float)half_two_v * pi);
    float x1 = cos(arg);
    float x2 = sin(arg);
    
    /* recursively compute DIT */
    if (npro_this_level/2 <= 1) {
      dit(input,v_minus,lambda,half_two_v,n,N_d,&real1,&imag1);
      dit(input,v_minus,lambda,half_two_v,(n+N_d/two_v)%N_d,N_d,&real2,&imag2);
    }
    else {
      npro_this_level = npro_this_level/2;
      if ((myrank < (npro_offset + npro_this_level/2))|| (npro_this_level <= 1)) {
	npro_offset = npro_offset;
	dit(input,v_minus,lambda,half_two_v,n,N_d,&real1,&imag1);
	npro_offset = npro_offset;
	if ((1 < npro_this_level) && (myrank == npro_offset)) {
	  MPI_Recv(&real2, 1, MPI_FLOAT, (npro_offset + npro_this_level/2), 0, 
		   MPI_COMM_WORLD, &status);
	  MPI_Recv(&imag2, 1, MPI_FLOAT, (npro_offset + npro_this_level/2), 1, 
		   MPI_COMM_WORLD,&status);
	}
      }
      else if (((npro_offset + npro_this_level/2) <= myrank) || (npro_this_level <= 1)){
	npro_offset = (npro_offset + npro_this_level/2);
	dit(input,v_minus,lambda,half_two_v,(n+N_d/two_v)%N_d,N_d,&real2,&imag2);
	npro_offset = (npro_offset - npro_this_level/2);
	if ((1 < npro_this_level) && (myrank == npro_offset + npro_this_level/2)) {
	  MPI_Send(&real2, 1, MPI_FLOAT, npro_offset, 0, MPI_COMM_WORLD);
	  MPI_Send(&imag2, 1, MPI_FLOAT, npro_offset, 1, MPI_COMM_WORLD);
	}
      }
      npro_this_level = npro_this_level*2;
    }
    
    
    /* do complex multiplication and store results */
    real3 = (real2 * x1) + (imag2 * x2);
    imag3 = (imag2 * x1) - (real2 * x2);
    *real_result_ptr = real1 - real3;
    *imag_result_ptr = imag1 - imag3;
  }
}



/* subroutine to calculate the dft using the decimation-in-time-frequency */
/* algorithm described by Ali Saidi in the ICASSP Proceedings of 1994. */

/* input: input - array of sample points */
/*   		 v - stage currently in */
/*			 gamma - node of the stage */
/*			 two_v - 2^v */
/*			 n - sample point currently calculating */
/*			 N_d - number of samples */
/*			 real_result_ptr - pointer to the real part of the calculated result */
/*			 imag_result_ptr - pointer to the imag part of the calculated result */
/*			 end_stage - the stage at which to transition to the recursive DIT */

/* output: None */

/* description: this function uses a recursive call to itself to calculate the */
/* 	 coefficients of the dft at the gamma node of stage v for the nth */
/*	 sample point using the following recursion. */
/*         if gamma = v = 0 */
/*		g[gamma,v](n) = x(n) */
/*         if gamma < 2^(v-1) */
/*		g[gamma,v](n) = g[gamma,v-1](n) + */
/*		 g[gamma,v-1]((n+N_d/(2^v))%N_d) */
/*	 else */
/*        	 lambda = gamma - 2^(v-1) */
/*	         g[gamma,v](n) = (g[gamma,v-1](n) - */
/*		  g[gamma,v-1]((n+N_d/(2^v))%N_d))W^(n*2^(v-1)) */

/*	this is repeated until v = (log2(N_d))/2 */
/*	at that point the remaining iterations are carried out by the */
/*	recursive decimation-in-time algorithm and the result is */
/*	multiplied by W^(gamma*n) */

void ditf_calc(float* input, int v, int gamma, int two_v, int n, int N_d,
	       float* real_result_ptr,	float* imag_result_ptr, int end_stage)
{
  
  /* local variables */
  
  
  float real1 = 0.0,
    real2 = 0.0,
    imag1 = 0.0,
    imag2 = 0.0;
  float x1,x2;
  
  
  /* determine if it is time to switch to the DIT portion of the recursion */
  
  if (v > end_stage) {
    
    /*  assign variables which are used multiple times to avoid recalculation */
    
    int v_minus = v - 1;
    int half_two_v = two_v/2;
    
    /* if v = gamma = 0, return x[n] */
    
    if ((v == 0) && (gamma == 0)) {
      *real_result_ptr = input[n];
      *imag_result_ptr = (float) 0;
    }
    
    /* if gamma <= 2^(v-1) do recursion */
    
    else if (gamma < (half_two_v)) {
      if (npro_this_level/2 <= 1) {
	ditf_calc(input,v_minus,gamma,half_two_v,n,N_d,&real1,&imag1,end_stage);
	ditf_calc(input,v_minus,gamma,half_two_v,(n+N_d/(two_v))%N_d,N_d,
		  &real2,&imag2,end_stage);
      }
      else {
	npro_this_level = npro_this_level/2;
	if ((myrank < (npro_offset + npro_this_level/2)) || (npro_this_level <= 1)) {
	  npro_offset = npro_offset;
	  ditf_calc(input,v_minus,gamma,half_two_v,n,N_d,&real1,&imag1,end_stage);	  
	  npro_offset = npro_offset;
	  if ((1 < npro_this_level) && (myrank == npro_offset)) {
	    MPI_Recv(&real2, 1, MPI_FLOAT, (npro_offset + npro_this_level/2), 0, 
		     MPI_COMM_WORLD, &status);
	    MPI_Recv(&imag2, 1, MPI_FLOAT, (npro_offset + npro_this_level/2), 1, 
		     MPI_COMM_WORLD,&status);
	  }
	}
	
	else if (((npro_offset + npro_this_level/2) <= myrank) || (npro_this_level <= 1)){
	  npro_offset = (npro_offset + npro_this_level/2);
	  ditf_calc(input,v_minus,gamma,half_two_v,(n+N_d/(two_v))%N_d,N_d,
		    &real2,&imag2,end_stage);
	  npro_offset = (npro_offset - npro_this_level/2);
	  if ((1 < npro_this_level) && (myrank == npro_offset + npro_this_level/2)) {
	    MPI_Send(&real2, 1, MPI_FLOAT, npro_offset, 0, MPI_COMM_WORLD);
	    MPI_Send(&imag2, 1, MPI_FLOAT, npro_offset, 1, MPI_COMM_WORLD);
	  }
	}
	npro_this_level = npro_this_level*2;
      }
      
      *real_result_ptr = real1 + real2;
      *imag_result_ptr = imag1 + imag2;
      
    }
    
    /* if gamma > 2^(v-1) do recursion */
    
    else {
      int lambda = gamma - half_two_v;
      float arg = ((n * half_two_v * 2 * pi) / N_d);
      if (npro_this_level/2 <= 1) {
	ditf_calc(input,v_minus,lambda,half_two_v,n,N_d,&real1,&imag1,end_stage);
	ditf_calc(input,v_minus,lambda,half_two_v,(n + N_d/(two_v))%N_d,N_d,
		  &real2,&imag2,end_stage);
      }
      else {
	npro_this_level = npro_this_level/2;
	if ((myrank < (npro_offset + npro_this_level/2))|| (npro_this_level <= 1)) {
	  npro_offset = npro_offset;
	  ditf_calc(input,v_minus,lambda,half_two_v,n,N_d,&real1,&imag1,end_stage);
	  npro_offset = npro_offset;
	  if ((1 < npro_this_level) && (myrank == npro_offset)) {
	    MPI_Recv(&real2, 1, MPI_FLOAT, (npro_offset + npro_this_level/2), 0, 
		     MPI_COMM_WORLD, &status);
	    MPI_Recv(&imag2, 1, MPI_FLOAT, (npro_offset + npro_this_level/2), 1, 
		     MPI_COMM_WORLD,&status);
	  }
	}
	else if (((npro_offset + npro_this_level/2) <= myrank) || (npro_this_level <= 1)){
	  npro_offset = (npro_offset + npro_this_level/2);
	  ditf_calc(input,v_minus,lambda,half_two_v,(n + N_d/(two_v))%N_d,N_d,
		    &real2,&imag2,end_stage);
	  npro_offset = (npro_offset - npro_this_level/2);
	  if ((1 < npro_this_level) && (myrank == npro_offset + npro_this_level/2)) {
	    MPI_Send(&real2, 1, MPI_FLOAT, npro_offset, 0, MPI_COMM_WORLD);
	    MPI_Send(&imag2, 1, MPI_FLOAT, npro_offset, 1, MPI_COMM_WORLD);
	  }
	}
	npro_this_level = npro_this_level*2;
      }
      real1 = real1-real2;
      imag1 = imag1-imag2;
      
      /* calculate W^(n*2^(v-1)) */
      
      x1 = cos(arg);
      x2 = sin(arg);
      
      /* do complex multiplication */
      
      *real_result_ptr = (real1 * x1) + (imag1 * x2);
      *imag_result_ptr = (imag1 * x1) - (real1 * x2);
      
    }
  }
  
  /* switch to the DIT portion of the recursion */
  
  else {
    
    /* calculate complex conversion factor */
    
    float arg = ((float)gamma * n * 2 * pi / (float)N_d);
    float x1 = cos(arg);
    float x2 = sin(arg);
    
    /* call the DIT recursion and multiply by conversion factor */
    
    dit(input,v,gamma,two_v,n,N_d,&real1,&imag1);
    *real_result_ptr = real1 * x1 + imag1 * x2;
    *imag_result_ptr = imag1 * x1 - real1 * x2;
  }
  
  
} /* end ditf_calc */
