/*// file: ditf.c*/
/*//*/

/*// include files*/
/*//*/
#include <stdio.h>
#include <math.h>
#include <mpi.h>


/*// define pi*/
/*//*/
#define pi 3.14159265359

/*// program: ditf*/
/*//*/
/*// arguments:*/
/*// 	N: number of samples*/
/*//		input: input array of N samples*/
/*//		real_a : array of size N to store real coeff's in.*/
/*//		imag_a : array of size N to store imag coeff's in.*/
/*

  return:
  logical value indicating status
*/
#define N 1024
#define Debug


int myrank;
int npro;

int main(int argc, char* argv[]) {
    int numb;
    float input[N];
    float real_a[N];
    float imag_a[N];
    float* local_real_a;
    float* local_imag_a;
    int i;
    void ditf_calc(float* input, int v, int gamma, int two_v, int n, int N_d,
		   float* real_result_ptr, float* imag_result_ptr,
		   int end_stage);
    void dit(float* input, int v, int gamma, int two_v, int n, int N_d,
	     float* real_result_ptr, float* imag_result_ptr);

    int N_d=N;
    int v=log10(N)/log10(2);	  /* assign v = log2(N_d)*/
    int two_v = N;		  /* two_v = 2^v */
    int end_stage = v/2;
    float temp_real_1, temp_imag_1;   /* dummy floats */
    int offset;

    MPI_Init(&argc , &argv);
    MPI_Comm_rank (MPI_COMM_WORLD,&myrank);
    MPI_Comm_size  (MPI_COMM_WORLD,&npro);

    numb = 2*(N_d/npro);
    offset = numb*myrank;
    if (myrank==0) {
	for (i=0;i<N_d;i++) {
	    input[i] = i;
	}
	input[0]=0;
	input[1]=1;
	input[2]=1;
	input[3]=0;
    }
/*    // Sends copy of input array to all processes*/
/*    //*/
    MPI_Bcast(input, N_d, MPI_FLOAT, 0, MPI_COMM_WORLD);

    local_real_a = (float*)malloc(numb * sizeof(float));
    local_imag_a = (float*)malloc(numb * sizeof(float));

    if ((myrank%2)==0) {
      for(i=offset; i<(offset+numb); i++) {
	ditf_calc(input,v,i,two_v,i,N_d,&local_real_a[i-offset],
		  &local_imag_a[i-offset],end_stage);
      }
    }
    else {
      for(i=offset; i<(offset+numb); i++) {
	ditf_calc(input,v,i,two_v,i,N_d,&temp_real_1,
		  &temp_imag_1,end_stage);
      }
    }
    MPI_Gather(local_real_a, numb, MPI_FLOAT, real_a, numb, MPI_FLOAT,
	       0,MPI_COMM_WORLD);
    MPI_Gather(local_imag_a, numb, MPI_FLOAT, imag_a, numb, MPI_FLOAT,
	       0,MPI_COMM_WORLD);


#ifdef Debug

    if(myrank == 0) {
	for(i=0;i<N_d;i++)
	    printf(" out[%d] = %f  +j %f \n",i,real_a[i],imag_a[i]);
    }

#endif


    MPI_Finalize();
/*    // exit gracefully*/
/*    //*/
    return 1;
    
}


/*// subroutine to calculate the dft using the recursive decimation-in-time*/
/*// algorithm described by Ali Saidi in the ICASSP Proceedings of 1994.*/
/*// this DIT should not be used alone because it outputs values in a special*/
/*// form to be consistent with the DITF*/
/*//*/
/*// input: input - array of sample points*/
/*//   		 v - stage currently in*/
/*//			 gamma - node of the stage*/
/*//			 two_v - 2^v*/
/*//			 n - sample point currently calculating*/
/*//			 N_d - number of samples*/
/*//			 real_result_ptr - pointer to the real part of the calculated result*/
/*//			 imag_result_ptr - pointer to the imag part of the calculated result*/
/*//*/
/*// output: None*/
/*//*/
/*// description: this function uses a recursive call to itself to calculate the*/
/*//		 coefficients of the dft at gamma node at stage v for the nth*/
/*//		 sample point using the following recursion.*/
/*//              if gamma = v = 0*/
/*//		       g'[gamma,v](n) = x(n)*/
/*//              if gamma < 2^(v-1)*/
/*//     	         g'[gamma,v](n) = g'[gamma,v-1](n) +
//		  g'[gamma,v-1]((n+N_d/(2^v))%N_d)W^(gamma/2^(v-1))
//		else
//		  lambda = gamma - 2^(v-1)
//		  g'[gamma,v](n) = g'[gamma,v-1](n) -
//		  g'[gamma,v-1]((n+N_d/(2^v))%N_d)W^(lambda/2^(v-1))
//
//*/
void dit(float* input, int v, int gamma, int two_v, int n, int N_d,
	 float* real_result_ptr,	float* imag_result_ptr)
{
    /* define local variables
     */
    float real1 = 0.0,
	real2 = 0.0,
	real3 = 0.0,
	imag1 = 0.0,
	imag2 = 0.0,
	imag3 = 0.0;
    
    /* define variables that will be used often to avoid repeat calculations
     */
    int v_minus = v-1;
    int half_two_v = two_v/2;
    
    /* if v = gamma = 0
    // assign return x[n]
    */
    if ((v == 0) && (gamma == 0)) {
	*real_result_ptr = input[n];
	*imag_result_ptr = (float) 0;
    }
    
    /* if gamma <= 2^(v-1) do recursion
     */
    else if (gamma < half_two_v) {
	
	/* calculate complex multiplier
	 */
	float arg = ((float)gamma / (float)half_two_v * pi);
	float x1 = cos(arg);
	float x2 = sin(arg);
	
	/* recursively compute the DIT
	 */
	dit(input,v_minus,gamma,half_two_v,n,N_d,&real1,&imag1);
	dit(input,v_minus,gamma,half_two_v,(n+N_d/two_v)%N_d,N_d,&real2,&imag2);
	
	/* do complex multiplication and store results
	 */
	real3 = (real2*x1) + (imag2*x2);
	imag3 = (imag2*x1) - (real2*x2);
	*real_result_ptr = real1 + real3;
	*imag_result_ptr = imag1 + imag3;
    }
    
    /* if gamma > 2^(v-1) then do recursion
     */
    else {
	
/*	// assign lambda = gamma - 2^(v-1)*/
/*	//*/
	int lambda = gamma - half_two_v;
	
/*	// compute complex multiplier*/
/*	//*/
	float arg = ((float)lambda / (float)half_two_v * pi);
	float x1 = cos(arg);
	float x2 = sin(arg);
	
/*	// recursively compute DIT*/
/*	//*/
	dit(input,v_minus,lambda,half_two_v,n,N_d,&real1,&imag1);
	dit(input,v_minus,lambda,half_two_v,(n+N_d/two_v)%N_d,N_d,&real2,&imag2);
	
/*	// do complex multiplication and store results*/
	real3 = (real2 * x1) + (imag2 * x2);
	imag3 = (imag2 * x1) - (real2 * x2);
	*real_result_ptr = real1 - real3;
	*imag_result_ptr = imag1 - imag3;
    }
}



/*// subroutine to calculate the dft using the decimation-in-time-frequency*/
/*// algorithm described by Ali Saidi in the ICASSP Proceedings of 1994.*/
/*//*/
/*// input: input - array of sample points*/
/*//   		 v - stage currently in*/
/*//			 gamma - node of the stage*/
/*//			 two_v - 2^v*/
/*//			 n - sample point currently calculating*/
/*//			 N_d - number of samples*/
/*//			 real_result_ptr - pointer to the real part of the calculated result*/
/*//			 imag_result_ptr - pointer to the imag part of the calculated result*/
/*//			 end_stage - the stage at which to transition to the recursive DIT*/
/*//*/
/*// output: None*/
/*//*/
/*// description: this function uses a recursive call to itself to calculate the*/
/*// 	 coefficients of the dft at the gamma node of stage v for the nth*/
/*//	 sample point using the following recursion.*/
/*//         if gamma = v = 0*/
/*//		g[gamma,v](n) = x(n)*/
/*//         if gamma < 2^(v-1)*/
/*//		g[gamma,v](n) = g[gamma,v-1](n) +*/
/*//		 g[gamma,v-1]((n+N_d/(2^v))%N_d)*/
/*//	 else*/
/*//        	 lambda = gamma - 2^(v-1)*/
/*//	         g[gamma,v](n) = (g[gamma,v-1](n) -*/
/*//		  g[gamma,v-1]((n+N_d/(2^v))%N_d))W^(n*2^(v-1))*/
/*//*/
/*//	this is repeated until v = (log2(N_d))/2*/
/*//	at that point the remaining iterations are carried out by the*/
/*//	recursive decimation-in-time algorithm and the result is*/
/*//	multiplied by W^(gamma*n)*/
/*//*/
void ditf_calc(float* input, int v, int gamma, int two_v, int n, int N_d,
	       float* real_result_ptr,	float* imag_result_ptr, int end_stage)
{
    
/*    // local variables*/
/*    //*/
    
    float real1 = 0.0,
	real2 = 0.0,
	imag1 = 0.0,
	imag2 = 0.0;
    float x1,x2;
    MPI_Status status;
    
    
/*    // determine if it is time to switch to the DIT portion of the recursion*/
/*    //*/
    if (v > end_stage) {
	
/*	//  assign variables which are used multiple times to avoid recalculation*/
/*	//*/
	int v_minus = v - 1;
	int half_two_v = two_v/2;
	
/*	// if v = gamma = 0, return x[n]*/
/*	//*/
	if ((v == 0) && (gamma == 0)) {
	    *real_result_ptr = input[n];
	    *imag_result_ptr = (float) 0;
	}
	
/*	// if gamma <= 2^(v-1) do recursion*/
/*	//*/
	else if (gamma < (half_two_v)) {
	  if ((myrank%2)==0) {
	    ditf_calc(input,v_minus,gamma,half_two_v,n,N_d,&real1,&imag1,end_stage);
	    MPI_Recv(&real2, 1, MPI_FLOAT, myrank+1, 0, MPI_COMM_WORLD, &status);
	    MPI_Recv(&imag2, 1, MPI_FLOAT, myrank+1, 1, MPI_COMM_WORLD, &status);
	  }
	  else if ((myrank%2)==1) {
	    ditf_calc(input,v_minus,gamma,half_two_v,(n+N_d/(two_v))%N_d,N_d,
		      &real2,&imag2,end_stage);
	    MPI_Send(&real2, 1, MPI_FLOAT, myrank-1, 0, MPI_COMM_WORLD);
	    MPI_Send(&imag2, 1, MPI_FLOAT, myrank-1, 1, MPI_COMM_WORLD);
	  }
	    *real_result_ptr = real1 + real2;
	    *imag_result_ptr = imag1 + imag2;
	}
	
/*	// if gamma > 2^(v-1) do recursion*/
/*	//*/
	else {
	    int lambda = gamma - half_two_v;
	    float arg = ((n * half_two_v * 2 * pi) / N_d);
	  if ((myrank%2)==0) {
	    ditf_calc(input,v_minus,lambda,half_two_v,n,N_d,&real1,&imag1,end_stage);
	    MPI_Recv(&real2, 1, MPI_FLOAT, myrank+1, 0, MPI_COMM_WORLD, &status);
	    MPI_Recv(&imag2, 1, MPI_FLOAT, myrank+1, 1, MPI_COMM_WORLD, &status);
	  }
	  else if ((myrank%2)==1) {
	    ditf_calc(input,v_minus,lambda,half_two_v,(n + N_d/(two_v))%N_d,N_d,
		      &real2,&imag2,end_stage;)
	    MPI_Send(&real2, 1, MPI_FLOAT, myrank-1, 0, MPI_COMM_WORLD);
	    MPI_Send(&imag2, 1, MPI_FLOAT, myrank-1, 1, MPI_COMM_WORLD);
	  }



	    real1 = real1-real2;
	    imag1 = imag1-imag2;
	    
/*	    // calculate W^(n*2^(v-1))*/
/*	    //*/
	    x1 = cos(arg);
	    x2 = sin(arg);
	    
/*	    // do complex multiplication*/
/*	    //*/
	    *real_result_ptr = (real1 * x1) + (imag1 * x2);
	    *imag_result_ptr = (imag1 * x1) - (real1 * x2);
	    
	}
    }
    
/*    // switch to the DIT portion of the recursion*/
/*    //*/
    else {
	
/*	// calculate complex conversion factor*/
/*	//*/
	float arg = ((float)gamma * n * 2 * pi / (float)N_d);
	float x1 = cos(arg);
	float x2 = sin(arg);
	
/*	call the DIT recursion and multiply by conversion factor*/


	dit(input,v,gamma,two_v,n,N_d,&real1,&imag1);
	*real_result_ptr = real1 * x1 + imag1 * x2;
	*imag_result_ptr = imag1 * x1 - real1 * x2;
    }
    
    
} /* end ditf_calc*/









