/* file:  */


/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#define GOOD_TRIG
#include "./trigtbl.h"
#include "event.h"
#include "event_constants.h"

/* method: fht_real_cc
//
// This code is based on the algorithm as described in the work by
// Ron Mayer whose code is available as a free-ware. The algorithm is
// described in the paper Hsieh S. Hou,"The Fast Hartley Transform Algorithm",
// IEEE Transactions on Computers, February 1987.pp. 147-153
//
//
// arguments:
// float* output_a :(output) output data array 
//			length of output array will always be 2*N
//                      output data memory should be allocated by the
//			calling program
// float* input_a : (input) input data array
//			length of input array is N
//			input data memory should be allocated by the
//			calling program.
//
// return: a logical_1 value indicating status
//
// implements a real Fast Hartley transform 
*/

/* ONE BIG HACK NOTES:
 * this code generates its own input signal in memory it`s allocated
 * performs a parallel FHT. It then computes the Fourier Coeffiecients
 * from the FHT coefficients
 */



#define ISIP_TRUE 1
#define ISIP_FALSE 0
#define NODE0 0
#define NODE1 1
#define MSG_TAG 50
#define SQRT2   2*0.70710678118654752440084436210484
#define sizeOfFloat sizeof( float )

int myrank;
int npro;
MPI_Status status;

int main(int argc, char* argv[]) {
  
  int merge(int npro, FILE* event_file, char* file_string);
  int sort( char* file_string);

  double start_time;
  double end_time;
  FILE* event_file;
  FILE* local_event_file;
  char file_string[MAX_STRING_LENGTH];
  char local_file_string[MAX_STRING_LENGTH];
  
  float *input_a;
  float *temp_input;
  float *output_a;

  int i,j,kk;
  int k,k1,k2,k3,k4,kx;
  float *fi,*fn,*gi,*ftemp;
  float s1,c1,s2,c2,s3,c3,s4,c4,g0,f0,f1,g1,f2,g2,f3,g3,a,b;

/*  TRIG_VARS; */
  int t_lam;
  int InitValue;
  int ierror;
  
  int N_d;
  int N_d_each;

  /*
   * Perform  Initializations : establish comm, get rank, and
   * get number of process
   */
  MPI_Init( &argc , &argv );
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  MPI_Comm_size( MPI_COMM_WORLD, &npro );

  if( argc != 3) {
    if( myrank == 0 ) {
      printf("USAGE: %s <number_of_points> <event file name>\n",argv[0] );
    }
    MPI_Finalize();
    return( ISIP_FALSE );
  }

  strcpy(&file_string, argv[2]);
  sprintf(local_file_string, "%s.%d", file_string, myrank);
  local_event_file = fopen(local_file_string,"w");
  
  if ( myrank == 0 ) {
    event_file = fopen(file_string,"w");
  }

  /*
   * retrieve number of data points to use from command line
   */
  N_d = atoi( argv[1] );

  /*
   *determine number of pts each process will compute
   */
  N_d_each = N_d / npro;

  /*
   * allocate memory for input, ouput, and local storage
   */
  if( myrank == 0 ) {
    input_a  = ( float * )malloc( N_d * sizeOfFloat );
    output_a = ( float * )malloc( 2 * N_d * sizeOfFloat );
  }
  temp_input = ( float * )malloc( N_d * sizeOfFloat );
  
    
    
  /*
   * node 0 generates an input signal
   *   - This will be removed for acutal signal processing
   */
  if( myrank == 0)  {
    for( i = 0; i < N_d; i++ ) {
      input_a[ i ] = i;
    }
  }

  /* mark the initial time of each process */
  MPI_Barrier( MPI_COMM_WORLD );
  start_time = MPI_Wtime();  


  /*
   * node 0 will make a copy of the input signal
   */
  if( myrank == 0 ) {
    for( i = 0; i < N_d; i ++ )
      temp_input[i] = input_a[i];

 
  /*
   * prescramble input
   */
    for (k1=1,k2=0;k1<N_d;k1++) {
      for (k=N_d>>1; (!((k2^=k)&k)); k>>=1);
      if (k1>k2) {
        a=temp_input[k1];temp_input[k1]=temp_input[k2];temp_input[k2]=a;
      }
    }
  }
  
  /*
   * Send a copy of the data to other nodes
   */
  if( npro > 0 ) {
    if ( myrank == 0 ) {
      fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
	      MPI_Wtime()-start_time, myrank, ALL_NODES);
    }
    else {
      fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
	      MPI_Wtime()-start_time, myrank,NODE0);
    }

    MPI_Bcast( temp_input, N_d, MPI_FLOAT, NODE0,
               MPI_COMM_WORLD );

    fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	    MPI_Wtime()-start_time, myrank, myrank);
  }

  /*
   * set InitValue for Trig Generator
   */
  
  if(( myrank & 0x01 ) == 0 )
    InitValue =  0;
  else
    InitValue = -1;
 
  /*
   * k = log2(N_d)
   */
  for ( k=0 ; (1<<k)<N_d ; k++ );

  k  &= 0x01;
 
  /*
   * each process initializes pointers to the beginning and end
   * of its section of data
   */
  fi = ftemp = temp_input + myrank * N_d_each;
  fn = fi + N_d_each;
  if( k == 0 ) {
      for( ; fi < fn; fi += 4 ) {
        f1     = fi[0 ]-fi[1 ];
        f0     = fi[0 ]+fi[1 ];
        f3     = fi[2 ]-fi[3 ];
        f2     = fi[2 ]+fi[3 ];
        fi[2 ] = f0-f2;	
        fi[0 ] = f0+f2;
        fi[3 ] = f1-f3;	
        fi[1 ] = f1+f3;
      }
  }
  else {
    for( gi = fi + 1; fi < fn; fi += 8, gi +=8 )  {
      c1     = fi[0 ] - gi[0 ];
      s1     = fi[0 ] + gi[0 ];
      c2     = fi[2 ] - gi[2 ];
      s2     = fi[2 ] + gi[2 ];
      c3     = fi[4 ] - gi[4 ];
      s3     = fi[4 ] + gi[4 ];
      c4     = fi[6 ] - gi[6 ];
      s4     = fi[6 ] + gi[6 ];
      f1     = s1 - s2;	
      f0     = s1 + s2;
      g1     = c1 - c2;	
      g0     = c1 + c2;
      f3     = s3 - s4;	
      f2     = s3 + s4;
      g3     = SQRT2*c4;		
      g2     = SQRT2*c3;
      fi[4 ] = f0 - f2;
      fi[0 ] = f0 + f2;
      fi[6 ] = f1 - f3;
      fi[2 ] = f1 + f3;
      gi[4 ] = g0 - g2;
      gi[0 ] = g0 + g2;
      gi[6 ] = g1 - g3;
      gi[2 ] = g1 + g3;
    }
  }
  
  if( npro > 1 ) { 

    if ( myrank == 0 ) {
      fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
	      MPI_Wtime()-start_time, myrank, ALL_NODES);
    }
    else {
      fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
	      MPI_Wtime()-start_time, myrank, NODE0);
    }
      
    MPI_Gather( ftemp , N_d_each, MPI_FLOAT,
                temp_input, N_d_each, MPI_FLOAT, 
                NODE0, MPI_COMM_WORLD );
    fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	    MPI_Wtime()-start_time, myrank, myrank);
  } 
  
  if( N_d < 16 )  {
    if( npro > 1 ) {
      if ( myrank == 0 ) {
	end_time = MPI_Wtime()-start_time;
      }
	fprintf(local_event_file,"%d %f %d %d\n", STOP_EVENT, 
		MPI_Wtime()-start_time, myrank, myrank);
     

      fclose(local_event_file);

      if ( myrank == 0 ) {
	merge(npro, event_file, file_string);
	sort( file_string );
      }

      MPI_Finalize(); 
      return( ISIP_TRUE );
    }
  }
  
  /*
   * Send all of the nodes a new copy of the data
   */
  if( npro > 1) {
    if ( myrank == 0 ) {
      fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
	      MPI_Wtime()-start_time, myrank, ALL_NODES);
    }
    else {
      fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
	      MPI_Wtime()-start_time, myrank, NODE0);
    }
      
    MPI_Bcast( temp_input, N_d, MPI_FLOAT, NODE0,
               MPI_COMM_WORLD );

    fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	    MPI_Wtime()-start_time, myrank, myrank);
  }

  do  {
    k  += 2;
    k1  = 1  << k;
    k2  = k1 << 1;
    k4  = k2 << 1;
    k3  = k2 + k1;
    kx  = k1 >> 1;
    fi  = temp_input;
    gi  = fi + kx;
    fn  = temp_input + N_d;
    
    
    if( myrank == 0) {    /* if myrank is even, I have even data */
      if ( npro > 1 )
	  i = 2;              /* setup next loop for even data */
      else
	  i = 1;
      
      do {
        f1      = fi[0 ] - fi[k1];
        f0      = fi[0 ] + fi[k1];
        f3      = fi[k2] - fi[k3];
        f2      = fi[k2] + fi[k3];
        fi[k2]  = f0	  - f2;
        fi[0 ]  = f0	  + f2;
        fi[k3]  = f1	  - f3;
        fi[k1]  = f1	  + f3;
        g1      = gi[0 ] - gi[k1];
        g0      = gi[0 ] + gi[k1];
        g3      = SQRT2  * gi[k3];
        g2      = SQRT2  * gi[k2];
        gi[k2]  = g0	  - g2;
        gi[0 ]  = g0	  + g2;
        gi[k3]  = g1	  - g3;
        gi[k1]  = g1	  + g3;
        gi     += k4;
        fi     += k4;
      } while (fi<fn);
    }
    else
      i = 1;  /* setup next loop index for odd data */
    
    TRIG_INIT( k, c1, s1, InitValue );
    for( ;i<kx;i += 2 )    {
      TRIG_NEXT( k, c1, s1 );
      c2 = c1*c1 - s1*s1;
      s2 = 2*(c1*s1);
      fn = temp_input + N_d;
      fi = temp_input + i;
      gi = temp_input + k1-i;
      do {
        REAL a,b,g0,f0,f1,g1,f2,g2,f3,g3;
        b       = s2*fi[k1] - c2*gi[k1];
        a       = c2*fi[k1] + s2*gi[k1];
        f1      = fi[0 ]    - a;
        f0      = fi[0 ]    + a;
        g1      = gi[0 ]    - b;
        g0      = gi[0 ]    + b;
        b       = s2*fi[k3] - c2*gi[k3];
        a       = c2*fi[k3] + s2*gi[k3];
        f3      = fi[k2]    - a;
        f2      = fi[k2]    + a;
        g3      = gi[k2]    - b;
        g2      = gi[k2]    + b;
        b       = s1*f2     - c1*g3;
        a       = c1*f2     + s1*g3;
        fi[k2]  = f0        - a;
        fi[0 ]  = f0        + a;
        gi[k3]  = g1        - b;
        gi[k1]  = g1        + b;
        b       = c1*g2     - s1*f3;
        a       = s1*g2     + c1*f3;
        gi[k2]  = g0        - a;
        gi[0 ]  = g0        + a;
        fi[k3]  = f1        - b;
        fi[k1]  = f1        + b;
        gi     += k4;
        fi     += k4;
      } while( fi < fn );
    }
    TRIG_RESET( k, c1, s1 );
  } while( k4 < N_d );
  /*
   * node 0 sends a copy of the partially processed data to node 1
   */
  if( npro > 1 ) {
    if( myrank == 1 ) {
      fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
	      MPI_Wtime()-start_time, myrank, NODE0 );

      MPI_Send( temp_input, N_d,
		MPI_FLOAT, NODE0, MSG_TAG,
		MPI_COMM_WORLD);

      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank );
    }
    
    if( myrank == 0 ) {
      fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
	      MPI_Wtime()-start_time, NODE0, NODE1 );

      MPI_Recv( output_a, N_d,
                MPI_FLOAT, NODE1, MSG_TAG,
                MPI_COMM_WORLD, &status);

      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank );
    
      /*
       * copy odd index data into array with even data
       */
      for( i = 1; i < N_d; i += 2) {
	  temp_input[i] = output_a[i];
	  output_a[i] = 0.0;
      }
    }
  }

  if( myrank == 0 ) {
    a = b = 0;  
    
/*  if (algorithm_d == FT_FAST_HARTLEY_COMPLEX){
    
    // get the output data
    //
    for (kk = 0; kk < N_d; kk++) {
      output_a[kk] = temp_input[kk];
    }
  }*/
    
  
/*  if (algorithm_d == FT_FAST_HARTLEY_REAL){ */

   /*
    * get fourier coefficients from hartley coefficients
    */
    for (i = 1,j = N_d - 1, k = N_d/2; i < k; i = i + 1,j = j - 1) {
      a = temp_input[i];
      b = temp_input[j];
      temp_input[j] = (a - b) * 0.5;
      temp_input[i] = (a + b) * 0.5;
    }
    
    /*
     *interlace the output and put into output_a
     */
    output_a[0] = temp_input[0];
    output_a[1] = 0;
    
    for ( kk = 1; kk <= N_d/2; kk++){
	output_a[2*kk] = output_a[2*N_d - 2*kk ] = temp_input[kk];
	output_a[2*kk+1] = -temp_input[N_d - kk];
	output_a[2*N_d - 2*kk + 1 ] = temp_input[N_d - kk];
    }
    output_a[N_d+1] = 0.0;
    
  }
  if (myrank == 0) {

#ifdef PRINT_OUTPUT
      for( kk = 0; kk <= N_d - 1; kk = kk++ ) {
	  printf("output[%d] = %f   + j%f\n", kk,output_a[2*kk], output_a[2*kk+1]);
      }
#endif
  }

  if( myrank == 0 ) {
    free( input_a );
    free( output_a );  
  } 
  free( temp_input );

    fprintf(local_event_file,"%d %f %d %d\n", STOP_EVENT, 
	    MPI_Wtime()-start_time, myrank, myrank );

  fclose(local_event_file);      
  
  if ( myrank == 0 ) {
    merge(npro, event_file, file_string);
    sort( file_string );
  }

  MPI_Finalize();

  /* exit gracefully */
  return( ISIP_TRUE );
    
}
  



