#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include "event.h"
#include "event_constants.h"

#define NI 1
#define ISIP_TRUE 1
#define ISIP_FALSE 0

    FILE* local_event_file;
    double start_time;
    int myrank;

int main(int argc, char **argv) {


  void fft2_real_cc( float* img, float* input_a, int N_d, int Sign,
		     int myrank, int npro, MPI_Comm);

  int merge(int npro, FILE* event_file, char* file_string);
  int sort( char* file_string);

  double end_time;
  FILE* event_file;
  char file_string[MAX_STRING_LENGTH];
  char local_file_string[MAX_STRING_LENGTH];
    
    int npro,m,p, prev,next,numb,Arrsize;
    int i,j,k,N_d;
    float *real, *img;
    float *local_real, *local_img;
    int iter;
    
    /* To Initialize MPI */
    
    MPI_Init(&argc , &argv);
    MPI_Comm_rank (MPI_COMM_WORLD,&myrank);
    MPI_Comm_size  (MPI_COMM_WORLD,&npro);

    if( argc != 3) {
      if( myrank == 0 ) {
	printf("USAGE: %s <number_of_points> <event file name>\n",argv[0] );
      }
      MPI_Finalize();
      return( ISIP_FALSE );
    }    

    strcpy(&file_string, argv[2]);
    sprintf(local_file_string, "%s.%d", file_string, myrank);
    local_event_file = fopen(local_file_string,"w");

    if ( myrank == 0 ) {
	event_file = fopen(file_string,"w");
    }
    
    N_d = atoi(argv[1]);
    Arrsize = N_d;
    
    m = Arrsize  % npro;
            
    /* Gives number of elements that have to be sent to each processor.
     */
    numb = Arrsize/npro; 
    
    if( myrank ==0) {
	real = (float *)malloc (N_d *sizeof (float));       
	img = (float *)malloc (N_d *sizeof (float));       
	
	for(i=0; i<N_d; i++){
	    real[i] = (float)(i);
	    img[i] = 0.0;
	}
    }    
    
    MPI_Barrier( MPI_COMM_WORLD );
    start_time= MPI_Wtime();

    local_real = (float *)malloc (numb * sizeof (float)); 
    local_img = (float *)malloc (numb * sizeof (float));

    if( npro > 1 ) {
      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank,0);
      }
    
      MPI_Scatter(real,numb,MPI_FLOAT,local_real,numb,MPI_FLOAT,
		  0,MPI_COMM_WORLD);

      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    }

    if( npro > 1 ) {
      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank,0);
      }

      MPI_Scatter(img,numb,MPI_FLOAT,local_img,numb,MPI_FLOAT,
		  0,MPI_COMM_WORLD);

      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    }

        
    fft2_real_cc(local_real,local_img,Arrsize,1,myrank,npro,MPI_COMM_WORLD);
    
    if( npro > 1 ) { 

      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, 0);
      }
      
      MPI_Gather(local_real, numb, MPI_FLOAT, real, numb, MPI_FLOAT,
		 0,MPI_COMM_WORLD);
      
      
      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    } 


    if( npro > 1 ) { 

      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, 0);
      }
      

      MPI_Gather(local_img, numb, MPI_FLOAT, img, numb, MPI_FLOAT,
		 0,MPI_COMM_WORLD);
      
      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    } 
    
    
#ifdef Test
    if(myrank == 0){
	Scramble(real,img,N_d);
	printf("\n In Test1\n");
	for(i=0; i<N_d; i++)
	    printf("real[%d] = %6.2f, img[%d] = %6.2f\n",i,real[i],i,img[i]);
    }
#endif

    fprintf(local_event_file,"%d %f %d %d\n", STOP_EVENT, 
	    MPI_Wtime()-start_time, myrank, myrank );
    
    fclose(local_event_file);      

    if ( myrank == 0 ) {
      merge(npro, event_file, file_string);
      sort( file_string );
    }
    
    MPI_Finalize();
 

}



/*  This method is a fortran code converted to fit our requirements.
    The code follows the exact structure of that in the book by
    Proakis and Manolakis " Digital Signal Processing " pp.
    
    arguments:
    float_4* output_a : output data array
    float_4* input_a : input data array
    return: a logical_1 value indicating status
    implements a real discrete fourier transform
    */

#define TWOPI 2* 3.1415926535897932384626434
#define ISIP_TRUE 1
#define ISIP_FALSE 0

void fft2_real_cc( float* output_a, float* input_a, int N_d, 
		  int Sign, int myrank, int npro, MPI_Comm comm)
{

    /* declare local variables */
    
    int p,m,l,i1,i2,numb,cp;
    int i,j,k,n1,n2,Procom;
    float real_twid,imag_twid;
    float tempr,tempi; 
    float temp = 0; 
    float *wr;
    float *wi;
    float *Tout;
    float *Tin;
    int Nodeinfo;
    MPI_Status status;

    numb = N_d/npro;
    n2 = N_d;

    /* No of communications required */
    p = powof2_cc(npro);
    
    wr = (float *)malloc ((N_d+1) *sizeof (float));	
    wi = (float *)malloc ((N_d+1) *sizeof (float));		
    
    Tout = (float *)malloc ((numb) *sizeof (float));	
    Tin  = (float *)malloc ((numb) *sizeof (float));		
    
    /*/ create lookup table */
    /* get the lookup values */    
    lookup_cc(wr,wi,N_d,Sign);
    
    m = powof2_cc(N_d);
    if(m == -1){
	printf("The data is not power of 2 --Exiting\n");
	MPI_Finalize();
	exit(1);
    }
    i2 = 1;
    for(i=0; i<p; i++){   
	
	for(cp =0; cp<numb; cp++){
	    Tout[cp] = output_a[cp];
	    Tin[cp] = input_a[cp];
	}
	
	Procom = npro >> (i+1);
	Nodeinfo = (Procom ^ myrank);

	if( npro > 1 ) {
	  if ( (myrank%2) == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, Nodeinfo);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank,Nodeinfo);
	  }

	  MPI_Sendrecv_replace(Tout,numb,MPI_FLOAT,Nodeinfo,0,
			       Nodeinfo,0,comm,&status);
	  fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		  MPI_Wtime()-start_time, myrank, myrank);
	

	  if ( (myrank%2) == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank,Nodeinfo);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, Nodeinfo);
	  }
	  
	  MPI_Sendrecv_replace(Tin,numb,MPI_FLOAT,Nodeinfo,0,
			       Nodeinfo,0,comm,&status);
	  fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		  MPI_Wtime()-start_time, myrank, myrank);
	}
     
	n2 = n2 >> 1;
	i1 = 0;
	i1 = ((myrank % (npro >> i)) % (npro >> (i+1))) * numb;
	
	for(j = 0; j < numb; j++){
	    real_twid = wr[i1];
	    imag_twid = wi[i1];
	    i1 = i1+i2;
	    
	    for(k = j; k < numb; k =k+numb)
	    {
		if( (myrank % (npro>>i)) >= (npro >> (i+1))){
		    tempi = Tin[k] - input_a[k];
		    tempr = Tout[k] - output_a[k];
		    output_a[k] = real_twid*tempr + imag_twid*tempi;
		    input_a[k] = real_twid*tempi - imag_twid*tempr;
		}
		else{
		    input_a[k] = input_a[k] + Tin[k];	    
		    output_a[k] = output_a[k] + Tout[k];
		}
		
	    }
	}
	i2 = i2 << 1;
    }
		
    for(i = p; i < m; i++){
	n1 = n2;
	n2 = n2 >> 1;
	i1 = 0;
	i2 = (int)((N_d)/n1);
	
	for(j = 0; j < n2; j++){
	    real_twid = wr[i1];
	    imag_twid = wi[i1];
	    i1 = i1+i2;
	    
	    for(k = j; k < numb; k =k+n1)
	    {
		
		l = (k+n2);
		tempr = output_a[k] - output_a[l];
		output_a[k] = output_a[k] + output_a[l];
		tempi = input_a[k] - input_a[l];
		input_a[k] = input_a[k] + input_a[l];
		output_a[l] = real_twid*tempr + imag_twid*tempi;
		input_a[l] = real_twid*tempi - imag_twid*tempr;
	    }
	}
    }
    
#ifdef Test1
    if(myrank == 0){
	printf("MY rank is %d\n",myrank);
    for(i=0; i<numb; i++)
	printf("in[%d] = %6.2f\n",i,input_a[i]);
    for(i=0; i<numb; i++)
	printf("real[%d] = %6.2f\n",i,output_a[i]);  
}
#endif
    /* delete memory*/
    free (Tout);
    free (Tin);
    free (wr);
    free (wi); 
}



