/*  This method is a fortran code converted to fit our requirements.
    The code follows the exact structure of that in the book by
    Proakis and Manolakis " Digital Signal Processing " pp.
    
    arguments:
    float_4* output_a : output data array
    float_4* input_a : input data array
    return: a logical_1 value indicating status
    implements a real discrete fourier transform
    */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include "event.h"
#include "event_constants.h"

#define TWOPI 2* 3.1415926535897932384626434
#define ISIP_TRUE 1
#define ISIP_FALSE 0

void fft2_real_cc( float* output_a, float* input_a, int N_d, 
		  int Sign, int myrank, int npro, MPI_Comm comm)
{
  int merge(int npro, FILE* event_file, char* file_string);
  int sort( char* file_string);

    /* declare local variables */
    
    int p,m,l,i1,i2,numb,cp;
    int i,j,k,n1,n2,Procom;
    float real_twid,imag_twid;
    float tempr,tempi; 
    float temp = 0; 
    float *wr;
    float *wi;
    float *Tout;
    float *Tin;
    int Nodeinfo;
    MPI_Status status;


    numb = N_d/npro;
    n2 = N_d;

    /* No of communications required */
    p = powof2_cc(npro);
    
    wr = (float *)malloc ((N_d+1) *sizeof (float));	
    wi = (float *)malloc ((N_d+1) *sizeof (float));		
    
    Tout = (float *)malloc ((numb) *sizeof (float));	
    Tin  = (float *)malloc ((numb) *sizeof (float));		
    
    /*/ create lookup table */
    /* get the lookup values */    
    lookup_cc(wr,wi,N_d,Sign);
    
    m = powof2_cc(N_d);
    if(m == -1){
	printf("The data is not power of 2 --Exiting\n");
	MPI_Finalize();
	exit(1);
    }
    i2 = 1;
    for(i=0; i<p; i++){   
	
	for(cp =0; cp<numb; cp++){
	    Tout[cp] = output_a[cp];
	    Tin[cp] = input_a[cp];
	}
	
	Procom = npro >> (i+1);
	Nodeinfo = (Procom ^ myrank);

	if( npro > 1 ) {
	  if ( (myrank%2) == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myank, Nodeinfo);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank,Nodeinfo);
	  }

	  MPI_Sendrecv_replace(Tout,numb,MPI_FLOAT,Nodeinfo,0,
			       Nodeinfo,0,comm,&status);
	  fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		  MPI_Wtime()-start_time, myrank, myrank);
	

	  if ( (myrank%2) == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank,Nodeinfo);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myank, Nodeinfo);
	  }
	  
	  MPI_Sendrecv_replace(Tin,numb,MPI_FLOAT,Nodeinfo,0,
			       Nodeinfo,0,comm,&status);
	  fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		  MPI_Wtime()-start_time, myrank, myrank);
	}
     
	n2 = n2 >> 1;
	i1 = 0;
	i1 = ((myrank % (npro >> i)) % (npro >> (i+1))) * numb;
	
	for(j = 0; j < numb; j++){
	    real_twid = wr[i1];
	    imag_twid = wi[i1];
	    i1 = i1+i2;
	    
	    for(k = j; k < numb; k =k+numb)
	    {
		if( (myrank % (npro>>i)) >= (npro >> (i+1))){
		    tempi = Tin[k] - input_a[k];
		    tempr = Tout[k] - output_a[k];
		    output_a[k] = real_twid*tempr + imag_twid*tempi;
		    input_a[k] = real_twid*tempi - imag_twid*tempr;
		}
		else{
		    input_a[k] = input_a[k] + Tin[k];	    
		    output_a[k] = output_a[k] + Tout[k];
		}
		
	    }
	}
	i2 = i2 << 1;
    }
		
    for(i = p; i < m; i++){
	n1 = n2;
	n2 = n2 >> 1;
	i1 = 0;
	i2 = (int)((N_d)/n1);
	
	for(j = 0; j < n2; j++){
	    real_twid = wr[i1];
	    imag_twid = wi[i1];
	    i1 = i1+i2;
	    
	    for(k = j; k < numb; k =k+n1)
	    {
		
		l = (k+n2);
		tempr = output_a[k] - output_a[l];
		output_a[k] = output_a[k] + output_a[l];
		tempi = input_a[k] - input_a[l];
		input_a[k] = input_a[k] + input_a[l];
		output_a[l] = real_twid*tempr + imag_twid*tempi;
		input_a[l] = real_twid*tempi - imag_twid*tempr;
	    }
	}
    }
    
#ifdef Test1
    if(myrank == 0){
	printf("MY rank is %d\n",myrank);
    for(i=0; i<numb; i++)
	printf("in[%d] = %6.2f\n",i,input_a[i]);
    for(i=0; i<numb; i++)
	printf("real[%d] = %6.2f\n",i,output_a[i]);  
}
#endif
    /* delete memory*/
    free (Tout);
    free (Tin);
    free (wr);
    free (wi); 
}
	
