#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include "event.h"
#include "event_constants.h"

#define ISIP_TRUE 1
#define ISIP_FALSE 0
#define NI 1
#define NODE0 0

    FILE* local_event_file;
    double start_time;


int main(int argc, char **argv)
{
  int merge(int npro, FILE* event_file, char* file_string);
  int sort( char* file_string);

  double end_time;
  FILE* event_file;
  char file_string[MAX_STRING_LENGTH];
  char local_file_string[MAX_STRING_LENGTH];

  void fft4_real_cc( float* img, float* input_a, int N_d, int myrank, 
		     int npro, MPI_Comm comm);

    int myrank,npro,m,p, prev,next,numb;
    int i,j,k,N_d;
    float *real, *img;
    float *local_real, *local_img;
    double starttime, total_time,maxtime;
    int iter;
    int test;
    /* To Initialize MPI */
    
    MPI_Init(&argc , &argv);
    MPI_Comm_rank (MPI_COMM_WORLD,&myrank);
    MPI_Comm_size  (MPI_COMM_WORLD,&npro);
    
    if( argc != 3) {
      if( myrank == 0 ) {
	printf("USAGE: %s <number_of_points> <event file name>\n",argv[0] );
      }
      MPI_Finalize();
      return( ISIP_FALSE );
    }    

    strcpy(&file_string, argv[2]);
    sprintf(local_file_string, "%s.%d", file_string, myrank);
    local_event_file = fopen(local_file_string,"w");

    if ( myrank == 0 ) {
	event_file = fopen(file_string,"w");
    }
    
    N_d = atoi(argv[1]);

    
    /* Size is broadcasted as all the other processors need to know 
       the size of the array being used. */
    
    /*   Fox also assume's the same */  
    m= N_d  % npro;
    
    if (m!= 0)
    { 
	if (myrank == 0)
	    printf("SORRY NOT EQUIPPED TO HANDLE THIS\n");
	MPI_Finalize();
	exit(1);
    }
    
                
    /* Gives number of elements that have to be sent to each processor.
     */
    numb = N_d/npro; 
    
    if(numb < 4){
	printf("The data set too small to be Parallelized \n");
	MPI_Finalize();
	exit(1);
    }
    real = (float *)malloc (N_d *sizeof (float));       
    img = (float *)malloc (N_d *sizeof (float));       
	
    
    for(iter=0; iter<NI; iter++){
    	if( myrank ==0)
	{
	    
	    for(i=0; i< N_d; i++){
		real[i] = (float)(i);
		img[i] = 0.0;
	    }
	}

	MPI_Barrier( MPI_COMM_WORLD );
	start_time= MPI_Wtime();
    if( npro > 1 ) {
      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank,0);
      }

	MPI_Bcast(real,N_d,MPI_FLOAT,0,MPI_COMM_WORLD);
	fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		MPI_Wtime()-start_time, myrank, myrank);
    }
    if( npro > 1 ) {
      if ( myrank == 0 ) {
	fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		MPI_Wtime()-start_time, myrank, ALL_NODES);
      }
      else {
	fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		MPI_Wtime()-start_time, myrank,0);
      }
	MPI_Bcast(img,N_d,MPI_FLOAT,0,MPI_COMM_WORLD);

      fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
	      MPI_Wtime()-start_time, myrank, myrank);
    }


	fft4_real_cc(real,img,N_d,myrank,npro,MPI_COMM_WORLD);
	
	
	if(myrank == 0) 
	    Scramble(real,img,N_d);
    }

#ifdef Test
    if(myrank == 0){
	
	printf("\n In Test1\n");
	for(i=0; i<N_d; i++)
	    printf("real[%d] = %6.2f, img[%d] = %6.2f\n",i,real[i],i,img[i]);
    }
#endif

    fprintf(local_event_file,"%d %f %d %d\n", STOP_EVENT, 
	    MPI_Wtime()-start_time, myrank, myrank );
    
    fclose(local_event_file);      

    if ( myrank == 0 ) {
      merge(npro, event_file, file_string);
      sort( file_string );
    }

    MPI_Finalize();
}


/*
// method: fft4_complex_cc
//
// This method performs a standard radix-4 decimation-in-frequency algorithm
// The code follows the exact structure of that in the book by
// Burrus and Parks "DFT/FFT and Convolution Algorithms". // remove me
//
// arguments:
// float_4* output_a :(output) output data array 
//			length of output array will always be 2*N
//			memory is allocated internally, not by the calling
//			program.
// float_4* input_a : (input) input data array
//			length of input array is N
//			input data memory should be allocated by the
//			calling program.
//
// return: a logical_1 value indicating status
//
*/


#define TWOPI 2.0*3.1415926535897932384626434

void fft4_real_cc( float *output_a, float *temp_input, int N_d, int myrank, 
                  int npro, MPI_Comm comm)
{
    
    /* declare local variables
     */
    int m,i1,i2,i3,p,numb,cp,Procom,Counter;
    int n1,n2;
    float a,b,c,e,r1,r2,r3,r4,s1,s2,s3,s4,co1,co2,co3,si1,si2,si3;
    int kk,k,j,i,t,index=0;
    float temp = 0;
    float *Tout;
    float *Tin;
    int Nodeinfo;
    MPI_Status status;
    float *INARR, *OUTARR;
   
    numb = N_d/npro;
    n2 = N_d;

    Tout = (float *)malloc ((N_d) *sizeof (float));    
    Tin  = (float *)malloc ((N_d) *sizeof (float));   
    
   /* INARR = (float *)malloc ((N_d*N_d) *sizeof (float));    
    OUTARR  = (float *)malloc ((N_d*N_d) *sizeof (float));   
*/
    m = powof2_cc(N_d);
    p = powof2_cc(N_d);
    t = m%2;;
 
    if(t != 0 ){
        printf("The data is not power of 4 --Exiting\n");
        MPI_Finalize();
        exit(1);
    }

    for(k = 0; k < m/2; k++){
	n1 = n2;
	n2 = n2>>2;
	e = TWOPI/(float)n1;
	a = 0.0;
	Counter = 0;
	for(j = 0; j < n2; j++){
	    b = a+a;
	    c = a+b;
	    co1 = cos(a);
	    co2 = cos(b);
	    co3 = cos(c);
	    si1 = sin(a);
	    si2 = sin(b);
	    si3 = sin(c);
	    a = (float)(j+1)*e;
	    for(i = j; i < N_d; i = i+n1){
		if ((Counter % npro) == myrank){
		    i1 = i+ n2;
		    i2 = i1 + n2;
		    i3 = i2+n2;
/*		    printf(" j= %d i= %d, i1= %d, i2= %d, i3= %d, n1= %d,n2 =%d, myrank = %d count=%d\n",j,i,i1,i2 ,i3,n1,n2,myrank, Counter);*/
		    r1 = output_a[i]+output_a[i2];
		    r3 = output_a[i]-output_a[i2];
		    s1 = temp_input[i]+temp_input[i2];
		    s3 = temp_input[i]-temp_input[i2];
		    r2 = output_a[i1]+output_a[i3];
		    r4 = output_a[i1]-output_a[i3];
		    s2 = temp_input[i1]+temp_input[i3];
		    s4 = temp_input[i1]-temp_input[i3];
		    Tout[i] = output_a[i] = r1+r2;
		    r2 = r1-r2;      
		    r1 = r3-s4;
		    r3 = r3+s4;      
		    Tin[i] = temp_input[i] = s1+s2;
		    s2 = s1-s2;      
		    s1 = s3+r4;
		    s3 = s3-r4;      
		    
		    Tout[i1] = output_a[i1] = r3*co1 + s3*si1;
		    Tin[i1] = temp_input[i1] = s3*co1 - r3*si1;
		    Tout[i2] = output_a[i2] = r2*co2 + s2*si2;
		    Tin[i2] = temp_input[i2] = s2*co2 - r2*si2;
		    Tout[i3] = output_a[i3] = r1*co3 + s1*si3;
		    Tin[i3] = temp_input[i3] = s1*co3 - r1*si3;
		}
		Counter++;
	    }
	}

	if( npro > 1 ) { 

	  if ( myrank == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank, ALL_NODES);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, NODE0);
	  }
	  
	  MPI_Allreduce(Tout,output_a, N_d, MPI_FLOAT,MPI_SUM,comm);	

	  if ( myrank == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, ALL_NODES);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank, NODE0);
	  }

	  fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		  MPI_Wtime()-start_time, myrank, myrank);
	} 

	if( npro > 1 ) { 

	  if ( myrank == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank, ALL_NODES);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, NODE0);
	  }
	  

	  MPI_Allreduce(Tin,temp_input,N_d, MPI_FLOAT,MPI_SUM,comm);

	  if ( myrank == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, ALL_NODES);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank, NODE0);
	  }


	  fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		  MPI_Wtime()-start_time, myrank, myrank);
	} 

	/* DO AN ALL GATHER HERE */
    	for(cp =0; cp < N_d ; cp++){
	    Tout[cp] = 0.0;
	    Tin[cp] = 0.0;
	}
	

    }
}










