/*
// method: fft4_complex_cc
//
// This method performs a standard radix-4 decimation-in-frequency algorithm
// The code follows the exact structure of that in the book by
// Burrus and Parks "DFT/FFT and Convolution Algorithms". // remove me
//
// arguments:
// float_4* output_a :(output) output data array 
//			length of output array will always be 2*N
//			memory is allocated internally, not by the calling
//			program.
// float_4* input_a : (input) input data array
//			length of input array is N
//			input data memory should be allocated by the
//			calling program.
//
// return: a logical_1 value indicating status
//
*/


#include <math.h>
#include <mpi.h>
#define TWOPI 2.0*3.1415926535897932384626434

void fft4_real_cc( float *output_a, float *temp_input, int N_d, int myrank, 
                  int npro, MPI_Comm comm)
{
    
    /* declare local variables
     */
    int m,i1,i2,i3,p,numb,cp,Procom,Counter;
    int n1,n2;
    float a,b,c,e,r1,r2,r3,r4,s1,s2,s3,s4,co1,co2,co3,si1,si2,si3;
    int kk,k,j,i,t,index=0;
    float temp = 0;
    float *Tout;
    float *Tin;
    int Nodeinfo;
    MPI_Status status;
    float *INARR, *OUTARR;
   
    numb = N_d/npro;
    n2 = N_d;

    Tout = (float *)calloc ((N_d) *sizeof (float));    
    Tin  = (float *)calloc ((N_d) *sizeof (float));   
    
   /* INARR = (float *)malloc ((N_d*N_d) *sizeof (float));    
    OUTARR  = (float *)malloc ((N_d*N_d) *sizeof (float));   
*/
    m = powof2_cc(N_d);
    p = powof2_cc(N_d);
    t = m%2;;
 
    if(t != 0 ){
        printf("The data is not power of 4 --Exiting\n");
        MPI_Finalize();
        exit(1);
    }

    for(k = 0; k < m/2; k++){
	n1 = n2;
	n2 = n2>>2;
	e = TWOPI/(float)n1;
	a = 0.0;
	Counter = 0;
	for(j = 0; j < n2; j++){
	    b = a+a;
	    c = a+b;
	    co1 = cos(a);
	    co2 = cos(b);
	    co3 = cos(c);
	    si1 = sin(a);
	    si2 = sin(b);
	    si3 = sin(c);
	    a = (float)(j+1)*e;
	    for(i = j; i < N_d; i = i+n1){
		if ((Counter % npro) == myrank){
		    i1 = i+ n2;
		    i2 = i1 + n2;
		    i3 = i2+n2;
/*		    printf(" j= %d i= %d, i1= %d, i2= %d, i3= %d, n1= %d,n2 =%d, myrank = %d count=%d\n",j,i,i1,i2 ,i3,n1,n2,myrank, Counter);*/
		    r1 = output_a[i]+output_a[i2];
		    r3 = output_a[i]-output_a[i2];
		    s1 = temp_input[i]+temp_input[i2];
		    s3 = temp_input[i]-temp_input[i2];
		    r2 = output_a[i1]+output_a[i3];
		    r4 = output_a[i1]-output_a[i3];
		    s2 = temp_input[i1]+temp_input[i3];
		    s4 = temp_input[i1]-temp_input[i3];
		    Tout[i] = output_a[i] = r1+r2;
		    r2 = r1-r2;      
		    r1 = r3-s4;
		    r3 = r3+s4;      
		    Tin[i] = temp_input[i] = s1+s2;
		    s2 = s1-s2;      
		    s1 = s3+r4;
		    s3 = s3-r4;      
		    
		    Tout[i1] = output_a[i1] = r3*co1 + s3*si1;
		    Tin[i1] = temp_input[i1] = s3*co1 - r3*si1;
		    Tout[i2] = output_a[i2] = r2*co2 + s2*si2;
		    Tin[i2] = temp_input[i2] = s2*co2 - r2*si2;
		    Tout[i3] = output_a[i3] = r1*co3 + s1*si3;
		    Tin[i3] = temp_input[i3] = s1*co3 - r1*si3;
		}
		Counter++;
	    }
	}

	MPI_Allreduce(Tout,output_a, N_d, MPI_FLOAT,MPI_SUM,comm);	
	MPI_Allreduce(Tin,temp_input,N_d, MPI_FLOAT,MPI_SUM,comm);

	/* DO AN ALL GATHER HERE */
    	for(cp =0; cp < N_d ; cp++){
	    Tout[cp] = 0.0;
	    Tin[cp] = 0.0;
	}
	

    }
}


