#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include "event.h"
#include "event_constants.h"

#define ISIP_TRUE 1
#define ISIP_FALSE 0
#define NI 1
#define NODE0 0
    FILE* local_event_file;
    double start_time;


int main(int argc, char **argv)
{
    void srfft_real_cc( float* img, float* input_a, int N_d, int myrank, 
			int npro, MPI_Comm comm);

  int merge(int npro, FILE* event_file, char* file_string);
  int sort( char* file_string);

  double end_time;
  FILE* event_file;
  char file_string[MAX_STRING_LENGTH];
  char local_file_string[MAX_STRING_LENGTH];


    int myrank,npro,m,p, prev,next,numb;
    int i,j,k,N_d;
    float *real, *img;
    float *local_real, *local_img;
    double starttime, total_time,maxtime;
    int iter;
    int test;
    /* To Initialize MPI */
    
    MPI_Init(&argc , &argv);
    MPI_Comm_rank (MPI_COMM_WORLD,&myrank);
    MPI_Comm_size  (MPI_COMM_WORLD,&npro);
    
    if( argc != 3) {
      if( myrank == 0 ) {
	printf("USAGE: %s <number_of_points> <event file name>\n",argv[0] );
      }
      MPI_Finalize();
      return( ISIP_FALSE );
    }    

    strcpy(&file_string, argv[2]);
    sprintf(local_file_string, "%s.%d", file_string, myrank);
    local_event_file = fopen(local_file_string,"w");

    if ( myrank == 0 ) {
	event_file = fopen(file_string,"w");
    }
    
    N_d = atoi(argv[1]);


    m= N_d  % npro;
    
    if (m!= 0)
    { 
	if (myrank == 0)
	    printf("SORRY NOT EQUIPPED TO HANDLE THIS\n");
	MPI_Finalize();
	exit(1);
    }
    
                
    /* Gives number of elements that have to be sent to each processor.
     */
    numb = N_d/npro; 
    
    if(numb < 2){
	printf("The data set too small to be Parallelized \n");
	MPI_Finalize();
	exit(1);
    }
    real = (float *)malloc (N_d *sizeof (float));       
    img = (float *)malloc (N_d *sizeof (float));       
	
    
    for(iter=0; iter<NI; iter++){
    	if( myrank ==0)
	{
	    
	    for(i=0; i< N_d; i++){
		real[i] = (float)(i);
		img[i] = 0.0;
	    }
	}

	MPI_Barrier( MPI_COMM_WORLD );
	start_time= MPI_Wtime();
	Scramble(real,N_d);
	if( npro > 1 ) {
	    if ( myrank == 0 ) {
		fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
			MPI_Wtime()-start_time, myrank, ALL_NODES);
	    }
	    else {
		fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
			MPI_Wtime()-start_time, myrank,0);
	    }

	    MPI_Bcast(real,N_d,MPI_FLOAT,0,MPI_COMM_WORLD);
	    fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		    MPI_Wtime()-start_time, myrank, myrank);
	}

	if( npro > 1 ) {
	    if ( myrank == 0 ) {
		fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
			MPI_Wtime()-start_time, myrank, ALL_NODES);
	    }
	    else {
		fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
			MPI_Wtime()-start_time, myrank,0);
	    }
	
	    MPI_Bcast(img,N_d,MPI_FLOAT,0,MPI_COMM_WORLD);
	    fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		    MPI_Wtime()-start_time, myrank, myrank);
	}

	srfft_real_cc(real,img,N_d,myrank,npro,MPI_COMM_WORLD);
	
	if (myrank ==0){
	    
	    /*  put into real and imaginary arrays
	     */
	    for (j = 1;j <= N_d/2;j++){
		img[j] = real[N_d - j];
		img[N_d - j] = -img[j];
		real[N_d - j] = real[j];
	    }
	    img[N_d/2] = 0.0;
	}
	
    }
    fprintf(local_event_file,"%d %f %d %d\n", STOP_EVENT, 
	    MPI_Wtime()-start_time, myrank, myrank );
    
    fclose(local_event_file);      

    if ( myrank == 0 ) {
      merge(npro, event_file, file_string);
      sort( file_string );
    }


    MPI_Finalize();
 

}


#define TWOPI  2* 3.1415926535897932384626434
#define DSQC    0.707106781186547f

void srfft_real_cc(float *real_a, float *imag_a, int N_d, int myrank,
		   int npro, MPI_Comm comm)
{
    
    float   t0, t1, t2, t3, t4, t5, t6, t;
    float   cc1, ss1, cc3, ss3;
    int     i, j, k, l, n1, n2, n4, n8, inc;
    int     is, id, i1, i2, i3, i4, i5, i6, i7, i8,m;
    int  Counter = 0, Counter1 =0 , cp;
    float *Treal;
    float *Timag;

    /*  // create sin-cos lookup table
	// */
    float* c = (float *)malloc (N_d *sizeof (float)); 
    float* s = (float *)malloc (N_d *sizeof (float)); 

    srfft_new_lookup_cc((float*)c,(float*)s, N_d);
    
    m = powof2_cc(N_d);
    
    /*   // bit reversal routine
	 // */
    /* Scramble(real_a,N_d); Bit reversal is done in the process zero 
       before doing a broad cast so don't have to do it here.*/
    /* Scramble done outside this routine and then data passed to
       this routine.*/
    /*  First (twiddleless) L Butterflies   
     */
    is = 0;
    id = 4;
    n1 = N_d - 1;
    Treal = (float *)malloc ((N_d) *sizeof (float));    
    Timag  = (float *)malloc ((N_d) *sizeof (float)); 

    do {
	for (i = is; i < n1; i = i + id){
	    t1 = real_a[i];
	    i1 = i + 1;
	    t2 = real_a[i1];
	    Treal[i] = real_a[i] = t1 + t2;
	    Treal[i] = real_a[i1] = t1 - t2;
	}
	is = (id << 1) - 2;
	id = id<<2;
    } while (is < n1) ;
    
    /*  Other (twiddled) L Butterflies  
     */
    n2 = 2;
    for (k = 2; k <= m; k++){
	n4 = n2>>1;
	n8 = n4>>1;
	n2 = n2<<1;
	is = 0;
	id = n2<<1;
	Counter =0;
	do {
	    for (i = is; i < N_d; i = i +id){
		if((Counter % npro ) == myrank){
		i2 = i + n4;
		i3 = i2 + n4;
		i4 = i3 + n4;
		t3 = real_a[i3];
		t4 = real_a[i4];
		t1 = real_a[i];
		t2 = t4 + t3;
		Treal[i4] = real_a[i4] = t4 - t3;
		Treal[i3] = real_a[i3] = t1 - t2;
		Treal[i] = real_a[i] = t1 + t2;
		
		if (n4 != 1)
		{
		    i1 = i + n8;
		    i2 += n8;
		    i3 += n8;
		    i4 += n8;
		    t3 = real_a[i3];
		    t4 = real_a[i4];
		    t1 = (t3 + t4)*DSQC;
		    t2 = (t3 - t4)*DSQC;
		    t3 = real_a[i1];
		    t4 = real_a[i2];
		    Treal[i4] = real_a[i4] = t4 - t1;
		    Treal[i3] = real_a[i3] = -t4 - t1;
		    Treal[i2] = real_a[i2] = t3 - t2;
		    Treal[i1] = real_a[i1] = t3 + t2;
		}
	    }
		Counter++;
	    }
	    is = (id << 1) - n2;
	    id = id << 2;
	} while (is < n1);

	if( npro > 1 ) { 

	  if ( myrank == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank, ALL_NODES);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, NODE0);
	  }

	  MPI_Allreduce(Treal,real_a, N_d, MPI_FLOAT,MPI_SUM,comm);
	  
	  if ( myrank == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, ALL_NODES);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank, NODE0);
	  }

	  fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		  MPI_Wtime()-start_time, myrank, myrank);
	} 
      
	    
	    for(cp =0; cp < N_d ; cp++){
		Treal[cp] = 0.0;
	    }
	    
	inc = N_d/n2;
	l = inc;
	for (j = 1; j < n8; j++)
	{
	    cc1 = c[l];
	    ss1 = s[l];
	    i1 = l*3;
	    cc3 = c[i1];
	ss3 = s[i1];
	l = (j + 1)*inc;
	is = 0;
	id = n2 << 1;
	Counter1 =  0;
	do
	  {
	    for (i = is; i < N_d; i += id)
	      {
		  if((Counter1 % npro == myrank)){
		i1 = i + j;
		i2 = i1 + n4;
		i3 = i2 + n4;
		i4 = i3 + n4;
		i5 = i - j + n4;
		i6 = i5 + n4;
		i7 = i6 + n4;
		i8 = i7 + n4;
		t3 = real_a[i3];
		t4 = real_a[i7];
		t1 = t3*cc1 + t4*ss1;
		t2 = t4*cc1 - t3*ss1;
		t5 = real_a[i4];
		t6 = real_a[i8];
		t3 = t5*cc3 + t6*ss3;
		t4 = t6*cc3 - t5*ss3;
		t5 = t1 + t3;
		t3 = t1 - t3;
		t6 = t2 + t4;
		t4 = t2 - t4;
		t1 = real_a[i6];
		Treal[i8] = real_a[i8] = t1 + t6;
		Treal[i3] =  real_a[i3] = t6 - t1;
 		t2 = real_a[i2];
		Treal[i4] = real_a[i4] = t2 - t3;
		Treal[i7] = real_a[i7] = -t2 - t3;
		t1 = real_a[i1];
		Treal[i1] = real_a[i1] = t1 + t5;
		real_a[i6] = t1 - t5;
		t5 = real_a[i5];
		Treal[i2] = real_a[i2] = t5 + t4;
		Treal[i5] = real_a[i5] = t5 - t4;
	      }
	      }
	    Counter1 ++;
	    is = (id << 1) - n2;
	    id = id << 2;
	  } while (is < n1);
	}
	if( npro > 1 ) { 

	  if ( myrank == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank, ALL_NODES);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, NODE0);
	  }
	  

	MPI_Allreduce(Treal,real_a, N_d, MPI_FLOAT,MPI_SUM,comm);


	  if ( myrank == 0 ) {
	    fprintf(local_event_file,"%d %f %d %d\n", SEND_EVENT, 
		    MPI_Wtime()-start_time, myrank, ALL_NODES);
	  }
	  else {
	    fprintf(local_event_file,"%d %f %d %d\n", RECV_EVENT, 
		    MPI_Wtime()-start_time, myrank, NODE0);
	  }


	  fprintf(local_event_file,"%d %f %d %d\n", WORKING_EVENT, 
		  MPI_Wtime()-start_time, myrank, myrank);
	} 

	for(cp =0; cp < N_d ; cp++){
	    Treal[cp] = 0.0;
	}

    }



  
  /* clear memory
  */
  free (c);
  free (s);
}



